--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Fuzzy.Abstract_Edit.Named                   Luebeck            --
--  Implementation                                 Summer, 2003       --
--                                                                    --
--                                Last revision :  16:12 16 Apr 2005  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Ada.IO_Exceptions;    use Ada.IO_Exceptions;
with Fuzzy.Basic_Edit;     use Fuzzy.Basic_Edit;
with Strings_Edit.Fields;  use Strings_Edit.Fields;

package body Fuzzy.Abstract_Edit.Named is

   procedure Erase (Domain : in out Domain_Description) is
   begin
      Domain.First := 1;
      Domain.Last  := 0;
      Erase (Domain.Name_Map);
   end Erase;

   procedure Copy
             (  Source : Domain_Description;
                Target : in out Domain_Description
             )  is
   begin
      Target.First    := Source.First;
      Target.Last     := Source.Last;
      Target.Name_Map := Source.Name_Map;
      for Index in Target.First..Target.Last loop
         Put (Target.Index_Map, Index, Get (Source.Index_Map, Index));
      end loop;
   end Copy;

   procedure Get
             (  Source  : String;
                Pointer : in out Integer;
                Domain  : Domain_Description;
                Index   : out Integer
             )  is
  begin
      Get (Source, Pointer, Domain.Name_Map, Index);
   end Get;
   
   function Get_Index 
            (  Domain : Domain_Description;
               Name   : String
            )  return Integer is
   begin
      return Find (Domain.Name_Map, Name);
   end Get_Index;

   function Get_Name
            (  Domain : Domain_Description;
               Index  : Integer
            )  return String is
   begin
      return GetName (Domain.Name_Map, Get (Domain.Index_Map, Index));
   exception
      when End_Error =>
         raise Constraint_Error;
   end Get_Name;

   procedure Get_Max_Range
             (  Domain  : Domain_Description;
                First   : out Integer;
                Last    : out Integer
             )  is
   begin
      First := Domain.First;
      Last  := Domain.Last;
   end Get_Max_Range;

   function Get_Cardinality (Domain : Domain_Description)
      return Natural is
   begin
      return Domain.Last - Domain.First + 1;
   end Get_Cardinality;

   procedure Get
             (  Source  : String;
                Pointer : in out Integer;
                Domain  : Domain_Description;
                From    : out Integer;
                To      : out Integer
             )  is
      Got_It : Boolean;
      First  : Integer;
   begin
      Get (Source, Pointer, Domain.Name_Map, First);
      Get_Delimiter (Source, Pointer, Got_It, Ellipsis);
      if Got_It then
         declare
            Last : Integer;
         begin
            Get (Source, Pointer, Domain.Name_Map, Last);
            if Last < First then
               raise Data_Error;
            end if;
            To := Last;
         exception
            when End_Error =>
               raise Data_Error;
         end;
      else
         To := First;
      end if;
      From := First;
   end Get;

   procedure Get
             (  Source  : String;
                Pointer : in out Integer;
                Domain  : Domain_Description;
                From    : out Value_Index;
                To      : out Value_Index
             )  is
      First : Integer;
      Last  : Integer;
   begin
      Get (Source, Pointer, Domain, First, Last);
      From := Value_Index (First - Domain.First + 1);
      To   := Value_Index (Last  - Domain.First + 1);
   end Get;

   procedure Put
             (  Destination : in out String;
                Pointer     : in out Integer;
                Domain      : Domain_Description;
                From        : Integer;
                To          : Integer;
                Field       : Natural   := 0;
                Justify     : Alignment := Left;
                Fill        : Character := ' '
             )  is
      Out_Field : constant Natural :=
                     Get_Output_Field (Destination, Pointer, Field);
      subtype Output is String (Pointer..Pointer + Out_Field - 1);
      Text  : Output renames
                 Destination (Pointer..Pointer + Out_Field - 1);
      Index : Integer := Pointer;
   begin
      if (  From > To
         or else
            From < Domain.First
         or else
            To > Domain.Last
         )
      then
         raise Constraint_Error;
      end if;
      Put
      (  Text,
         Index,
         GetName (Domain.Name_Map, Domain.Index_Map.Vector (From))
      );
      if From /= To then
         Put (Text, Index, "..");
         Put
         (  Text,
            Index,
            GetName (Domain.Name_Map, Domain.Index_Map.Vector (To))
         );
      end if;
      Adjust_Output_Field
      (  Destination,
         Pointer,
         Index,
         Out_Field,
         Field,
         Justify,
         Fill
      );
   exception
      when End_Error =>
         raise Constraint_Error;
   end Put;

   procedure Put
             (  Destination : in out String;
                Pointer     : in out Integer;
                Domain      : Domain_Description;
                From        : Value_Index;
                To          : Value_Index;
                Field       : Natural   := 0;
                Justify     : Alignment := Left;
                Fill        : Character := ' '
             )  is
   begin
      Put
      (  Destination => Destination,
         Pointer     => Pointer,
         Domain      => Domain,
         From        => Integer (From) + Domain.First - 1,
         To          => Integer (To)   + Domain.First - 1,
         Field       => Field,
         Justify     => Justify,
         Fill        => Fill
      );
   end Put;

   procedure Add
             (  Domain : in out Domain_Description;
                Name   : String;
                Index  : Integer
             )  is
   begin
      if Domain.First <= Domain.Last then
         if not
            (  (Index < Domain.First and then Index + 1 = Domain.First)
            or else
               (Index > Domain.Last and then Index - 1 = Domain.Last)
            )
         then
            raise Constraint_Error;
         end if;
      end if;
      Add (Domain.Name_Map, Name, Index);
      for Item in 1..GetSize (Domain.Name_Map) loop
         Put (Domain.Index_Map, GetTag (Domain.Name_Map, Item), Item);
      end loop;
      if Domain.First <= Domain.Last then
         if Domain.First > Index then
            Domain.First := Index;
         else
            Domain.Last  := Index;
         end if;
      else
         Domain.First := Index;
         Domain.Last  := Index;
      end if;
   end Add;

end Fuzzy.Abstract_Edit.Named;
