--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Fuzzy.Abstract_Edit.Named                   Luebeck            --
--  Interface                                      Summer, 2003       --
--                                                                    --
--                                Last revision :  16:12 16 Apr 2005  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
--  This package provides string I/O for fuzzy sets over discrete domain
--  sets,  with  named  elements.  The  type  Domain_Description  is   a
--  descendant  of  Fuzzy.Abstract_Edit.User_Data  which  provides   the
--  functionality. 
--
with Generic_Unbounded_Array;
with Name_Tables;

package Fuzzy.Abstract_Edit.Named is
   pragma Elaborate_Body (Fuzzy.Abstract_Edit.Named);
--
-- Domain_Description -- Discrete domain
--
   type Domain_Description is new User_Data with private;
--
-- Add -- Add one domain set element
--
--    Domain - To add an element
--    Name   - Of the element
--    Index  - Of the element (in a Set)
--
-- This procedure adds Name under Index to Domain.  Elements  should  be
-- added in an order keeping the set of  indices  contiguos.  Otherwise,
-- Constraint_Error is propagated. 
--
-- Exceptions :
--
--    Constraint_Error - Illegal name or index
--    Name_Error       - The name is already used
--
   procedure Add
             (  Domain : in out Domain_Description;
                Name   : String;
                Index  : Integer
             );
--
-- Copy -- A domain set
--
--    Source - To be copied
--    Target - The result
--
   procedure Copy
             (  Source : Domain_Description;
                Target : in out Domain_Description
             );
--
-- Erase -- Makes domain empty
--
--    Domain - To add an element
--
   procedure Erase (Domain : in out Domain_Description);
--
-- Get -- Replaces Fuzzy.Abstract_Edit...
--
   procedure Get
             (  Source  : String;
                Pointer : in out Integer;
                Domain  : Domain_Description;
                From    : out Integer;
                To      : out Integer
             );
--
-- Get -- Get an element
--
--    Source  - String to parse
--    Pointer - Starting string position
--    Domain  - The domain
--    Index   - Of the element
--
-- This procedure parses the string Source. If a name of an element from
-- Domain  is  successfully  matched  starting  from  Source  (Pointer),
-- Pointer is advanced to the position following the  matched  name  and
-- Index is set to respond to  the  domain  value  associated  with  the
-- variable.  
--
-- Exceptions :
--
--    End_Error    - Nothing matched
--    Layout_Error - Pointer is not in Source'First..Source'Last + 1
--
   procedure Get
             (  Source  : String;
                Pointer : in out Integer;
                Domain  : Domain_Description;
                Index   : out Integer
             );
   pragma Inline (Get);
--
-- Get_Cardinality -- Get number of elements in a domain
--
--    Domain - The domain
--
-- Returns :
--
--    The number of elements
--
   function Get_Cardinality (Domain : Domain_Description)
      return Natural;
   pragma Inline (Get_Cardinality);
--
-- Get_Index -- Of a domain set element
--
--    Domain - The domain description
--    Name   - Of the element
--
-- Returns :
--
--    The index of the element named by Name
--
-- Exceptions :
--
--    End_Error - No such element
--
   function Get_Index 
            (  Domain : Domain_Description;
               Name   : String
            )  return Integer;
   pragma Inline (Get_Index);
--
-- Get_Name -- Of a domain set element
--
--    Domain - The domain description
--    Index  - Of the element (in a Set)
--
-- Returns :
--
--    The name associated with Index
--
-- Exceptions :
--
--    Constraint_Error - No such element (illegal index)
--
   function Get_Name
            (  Domain : Domain_Description;
               Index  : Integer
            )  return String;
--
-- Get_Max_Range -- Overrides Fuzzy.Abstract_Edit...
--
   procedure Get_Max_Range
             (  Domain : Domain_Description;
                First  : out Integer;
                Last   : out Integer
             );
--
-- Put -- Replaces Fuzzy.Abstract_Edit...
--
   procedure Put
             (  Destination : in out String;
                Pointer     : in out Integer;
                Domain      : Domain_Description;
                From        : Integer;
                To          : Integer;
                Field       : Natural   := 0;
                Justify     : Alignment := Left;
                Fill        : Character := ' '
             );
private
   use Name_Tables.Name_Maps;

   type Positive_Array is array (Integer range <>) of Positive;
   package Set_To_Name_Maps is
      new Generic_Unbounded_Array
          (  Index_Type        => Integer,
             Object_Type       => Positive,
             Object_Array_Type => Positive_Array,
             Null_Element      => 1
          );
   use Set_To_Name_Maps;
--
-- Domain_Description -- Discrete domain
--
-- The  fields of the record: Name_Map is a table of names of the domain
-- set elements. Each element is associated with an  unique  index  from
-- First..Last.  The  sets  over  the  domain  are   declared   as   Set
-- (First..Last). Index_Map is an  array  which  for  every  index  from
-- First..Last contains  the  position  of  the  corresponding  name  in
-- Name_Map.   This   position   is   a   positive   number,  names  are
-- alphabetically ordered.
--
   type Domain_Description is new User_Data with record
      First     : Integer := 1;
      Last      : Integer := 0;
      Name_Map  : Dictionary;
      Index_Map : Unbounded_Array;
   end record;
--
-- Get -- Overrides Fuzzy.Abstract_Edit...
--
   procedure Get
             (  Source  : String;
                Pointer : in out Integer;
                Domain  : Domain_Description;
                From    : out Value_Index;
                To      : out Value_Index
             );
--
-- Put -- Overrides Fuzzy.Abstract_Edit...
--
   procedure Put
             (  Destination : in out String;
                Pointer     : in out Integer;
                Domain      : Domain_Description;
                From        : Value_Index;
                To          : Value_Index;
                Field       : Natural   := 0;
                Justify     : Alignment := Left;
                Fill        : Character := ' '
             );

end Fuzzy.Abstract_Edit.Named;
