--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Fuzzy.Linguistics.Measures                  Luebeck            --
--  Interface                                      Spring, 2005       --
--                                                                    --
--                                Last revision :  14:28 24 Jul 2005  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Units;   use Units;

with Fuzzy.Measures;
with Intervals.Measures;
with Measures;
with Units.Base;

generic
   with package Float_Measures is
      new Standard.Measures (Fuzzy_Floats.Number);
   with package Interval_Measures is
      new Intervals.Measures
          (  Float_Measures,
             Fuzzy_Floats.Float_Intervals
          );
   with package Fuzzy_Measures is
      new Fuzzy.Measures
          (  Fuzzy_Floats      => Fuzzy_Floats,
             Float_Measures    => Float_Measures,
             Interval_Measures => Interval_Measures
          );
package Fuzzy.Linguistics.Measures is
   use Float_Measures;
   use Interval_Measures;
   use Fuzzy_Measures;

   subtype Number is Fuzzy_Floats.Number;
--
-- Variable_Measure -- A dimensioned linguistic variable
--
--    SI     - The dimension unit
--    Gain   - The value of the variable in the dimension unit
--    Offset - The dimension unit shift
--
   type Variable_Measure (SI : Unit := Units.Base.Unitless) is record
      Gain   : Variable;
      Offset : Number'Base;
   end record;
--
-- Pair_Measure -- A measure - confidence level pair
--
   type Pair_Measure is record
      Value : Measure;
      Level : Confidence;
   end record;
--
-- Append -- One point to a linguistic variable
--
--    Var   - The variable
--    Value - The membership function argument
--    Level - The membership function value in the argument
--
-- This function is used to form the membership function of a linguistic
-- variable. It adds a new point to the  function.  Points  have  to  be
-- added in ascending order. It is allowed to define several points with
-- same argument (Value). The dimension units of Value  should  be  same
-- though  maybe  shifted  as ones of Var. Otherwise Constraint_Error is
-- propagated. 
--
-- Exceptions :
--
--    Data_Error - The point is out of order
--    Unit_Error - Incompatible units
--
   procedure Append
             (  Var   : in out Variable_Measure;
                Value : Measure;
                Level : Confidence
             );
--
-- Convert -- Measure conversion
--
--    Value - The variable measurement to convert
--    Scale - The measure of the result
--
-- This function is used  to convert a variable measure  Value  to  the
-- measurement units specified by the parameter Scale. When  offsets  of
-- Value and Scale are same this is the null operation. 
--
--    Convert (T, Celsius)  -- Temperature in Celsius degrees
--
-- Returns :
--
--    Scale equivalent of Value
--
-- Exceptions :
--
--    Unit_Error -- Item and Scale have different units
--
   function Convert (Value : Variable_Measure; Scale : Measure)
      return Variable_Measure;
   pragma Inline (Convert);   
--
-- Empty -- Empty variable
--
--    Scale - The measure of the result
--
-- Returns :
--
--    An empty variable
--
   function Empty (Scale : Measure) return Variable_Measure;
   pragma Inline (Empty);
--
-- Erase -- Makes a variable an empty set
--
--    Var - The variable
--
   procedure Erase (Var : in out Variable_Measure);
   pragma Inline (Erase);
--
-- Equal -- Compare two variables (equality)
--
--    A   - The first argument
--    B   - The second argument
--    Eps - Tolerance level
--
-- One of the arguments can be a value-confidence pair. Two  points  are
-- assumed same if their confidence levels are same and values differ no
-- more than Eps.
--
-- Returns :
--
--    True if A and B are exactly same
--
   function Equal (A, B : Variable_Measure; Eps : Number'Base := 0.0)
      return Boolean;
   function Equal
            (  A   : Variable_Measure;
               B   : Pair_Measure;
               Eps : Number'Base := 0.0
            )  return Boolean;
   function Equal
            (  A   : Pair_Measure;
               B   : Variable_Measure;
               Eps : Number'Base := 0.0
            )  return Boolean;
   function Equal
            (  A   : Pair_Measure;
               B   : Pair_Measure;
               Eps : Number'Base := 0.0
            )  return Boolean;
--
-- Get_Points_Number -- Get number of points of the membership function
--
--    Var - The variable
--
-- Returns :
--
--    The number of distinct points of the membership function
--
   function Get_Points_Number (Var : Variable_Measure) return Natural;
   pragma Inline (Get_Points_Number);
--
-- Get_Value -- Get SI value of a variable
--
--    Value - The variable
--
-- Returns :
--
--    SI equivalent of Value
--
   function Get_Value (Value : Variable_Measure) return Variable;
   pragma Inline (Get_Value);
--
-- Get_Value_As -- Get value measured in non-SI units
--
--    Value  - The variable
--    Scale  - The measure of the result
--
-- Returns :
--
--    Scale equivalent of Value in as a variable
--
-- Exceptions :
--
--    Unit_Error -- Value and Scale have different units
--
   function Get_Value_As (Value : Variable_Measure; Scale : Measure)
      return Variable;
   pragma Inline (Get_Value_As);
--
-- Get_Unit -- Get unit
--
--    Value  - The variable
--
-- Returns :
--
--    SI component
--
   function Get_Unit (Value : Variable_Measure) return Unit;
   pragma Inline (Get_Unit);   
--
-- Is_Empty -- Check if a variable is an empty set
--
--    Var - The variable
--
-- Returns :
--
--    True if the membership function is everywhere False
--
   function Is_Empty (Var : Variable_Measure) return Boolean;
--
-- Is_In -- Membership / subset test
--
--    A - Measure, interval/fuzzy/variable measure
--    B - Measure, interval/fuzzy/variable measure
--
-- At  least  one  argument  should be a variable measure. Arguments can
-- have any units. 
--
-- Returns :
--
--    (P(B|A), N(B|A))
--
   function Is_In (A, B : Variable_Measure) return Fuzzy_Boolean;
   function Is_In
            (  A : Measure;
               B : Variable_Measure
            )  return Fuzzy_Boolean;
   function Is_In
            (  A : Interval_Measure;
               B : Variable_Measure
            )  return Fuzzy_Boolean; 
   function Is_In
            (  A : Fuzzy_Measure;
               B : Variable_Measure
            )  return Fuzzy_Boolean;
   function Is_In
            (  A : Variable_Measure;
               B : Measure
            )  return Fuzzy_Boolean;
   function Is_In
            (  A : Variable_Measure;
               B : Interval_Measure
            )  return Fuzzy_Boolean;
   function Is_In
            (  A : Variable_Measure;
               B : Fuzzy_Measure
            )  return Fuzzy_Boolean;
--
-- Is_Interval -- Check if a variable is an interval
--
--    Var - The variable
--
-- This function returns True if the membership function of Var is 1  on
-- a finite interval and 0 outside it.
--
-- Returns :
--
--    True if it is an interval
--
   function Is_Interval (Var : Variable_Measure) return Boolean;
   pragma Inline (Is_Interval);
--
-- Is_Singleton -- Check if a variable is an singleton
--
--    Var   - The variable
--
-- This function returns True if the membership function of Vae is 1 in
-- one point and 0 elsewhere.
--
-- Returns :
--
--    True if is a singleton
--
   function Is_Singleton (Var : Variable_Measure) return Boolean;
   pragma Inline (Is_Singleton);
--
-- Necessity -- Conditional necessity
--
--    A - The first argument
--    B - The second argument (condition)
--
-- With  one  argument the condition is assumed to be the domain set. At
-- least one argument should be a variable.
--
-- Returns :
--
--    N(A|B)
--
   function Necessity (A : Variable_Measure) return Confidence;
   function Necessity (A, B : Variable_Measure) return Confidence;
   function Necessity
            (  A : Measure;
               B : Variable_Measure
            )  return Confidence;
   function Necessity
            (  A : Interval_Measure;
               B : Variable_Measure
            )  return Confidence;
   function Necessity
            (  A : Fuzzy_Measure;
               B : Variable_Measure
            )  return Confidence;
   function Necessity
            (  A : Variable_Measure;
               B : Measure
            )  return Confidence;
   function Necessity
            (  A : Variable_Measure;
               B : Interval_Measure
            )  return Confidence;
   function Necessity
            (  A : Variable_Measure;
               B : Fuzzy_Measure
            )  return Confidence;
--
-- Normalize -- Shift removing
--
--    Value - The variable
--
-- This  function  is  used  to  convert  Value  to  to  its   unshifted
-- equivalent.  
--
-- Returns :
--
--    Unshifted equivalent of Value
--
   function Normalize (Value : Variable_Measure)
      return Variable_Measure;
   pragma Inline (Normalize);
--
-- Possibility -- Conditional possibility
--
--    A - The first argument
--    B - The second argument (condition)
--
-- With  one  argument the condition is assumed to be the domain set. At
-- least one argument should be a variable. 
--
-- Returns :
--
--    P(A|B)
--
   function Possibility (A : Variable_Measure) return Confidence;
   function Possibility (A, B : Variable_Measure) return Confidence;
   function Possibility 
            (  A : Measure;
               B : Variable_Measure
            )  return Confidence;
   function Possibility 
            (  A : Interval_Measure;
               B : Variable_Measure
            )  return Confidence;
   function Possibility 
            (  A : Fuzzy_Measure;
               B : Variable_Measure
            )  return Confidence;
   function Possibility 
            (  A : Variable_Measure;
               B : Measure
            )  return Confidence;
   function Possibility 
            (  A : Variable_Measure;
               B : Interval_Measure
            )  return Confidence;
   function Possibility 
            (  A : Variable_Measure;
               B : Fuzzy_Measure
            )  return Confidence;
--
-- Shift -- Non-destructive shift
--
--    Value  - A variable to shift
--    Shift  - The shift
--
-- This function is used to convert Value to its shifted equivalent. 
--
-- Returns :
--
--    Shifted equivalent of Value
--
   function Shift (Value : Variable_Measure; Shift : Number'Base)
      return Variable_Measure;
   pragma Inline (Shift);
--
-- To_Interval_Measure -- Conversion
--
--    Value - A variable to convert
--
-- Returns :
--
--    The corresponding interval
--
-- Exceptions :
--
--    Constraint_Error - Value is not an interval
--
   function To_Interval_Measure (Value : Variable_Measure)
      return Interval_Measure;
   pragma Inline (To_Interval_Measure);
--
-- To_Measure -- Conversion
--
--    Value - A variable to convert
--
-- Returns :
--
--    The corresponding measure
--
-- Exceptions :
--
--    Constraint_Error - Value is not a singleton (number)
--
   function To_Measure (Value : Variable_Measure) return Measure;
   pragma Inline (To_Measure);
--
-- To_Variable_Measure -- Conversion
--
--    Value - A number, interval or a fuzzy number
--
-- Returns :
--
--    The corresponding variable
--
   function To_Variable_Measure (Value : Number)
      return Variable_Measure;
   function To_Variable_Measure (Value : Interval)
      return Variable_Measure;
   function To_Variable_Measure (Value : Fuzzy_Float)
      return Variable_Measure;
   function To_Variable_Measure (Value : Variable)
      return Variable_Measure;
   function To_Variable_Measure (Value : Measure)
      return Variable_Measure;
   function To_Variable_Measure (Value : Interval_Measure)
      return Variable_Measure;
   function To_Variable_Measure (Value : Fuzzy_Measure)
      return Variable_Measure;
   pragma Inline (To_Variable_Measure);
--
-- +, -, *, / -- By a scalar
--
--    Left  - Variable
--    Right - Number
--
-- The scalar in a multiplicative operation may be either a number or  a
-- dimensioned number. 
--
-- Returns :
--
--    The result of the operation
--
-- Exceptions :
--
--    Constraint_Error - Unit power overflow
--    Unit_Error       - Incompatible units
--
   function "+" (Left : Variable_Measure; Right : Measure)
      return Variable_Measure;
   function "-" (Left : Variable_Measure; Right : Measure)
      return Variable_Measure;
   function "*" (Left : Variable_Measure; Right : Number)
      return Variable_Measure;
   function "*" (Left : Variable_Measure; Right : Measure)
      return Variable_Measure;
   function "*" (Left : Variable; Right : Measure)
      return Variable_Measure;
   function "/" (Left : Variable_Measure; Right : Number)
      return Variable_Measure;
   function "/" (Left : Variable_Measure; Right : Measure)
      return Variable_Measure;
   function "/" (Left : Variable; Right : Measure)
      return Variable_Measure;
--
-- = -- Equality
--
--   A - Dimensioned number, interval, fuzzy number, variable
--   B - Dimensioned number, interval, fuzzy number, variable
--
-- At least one argument should be a dimensioned variable.
--
-- Returns :
--
--    (P(A|B), N(A|B) and N(B|A))
--
   function "=" (A, B : Variable_Measure) return Fuzzy_Boolean;
   function "=" (A : Measure; B : Variable_Measure)
      return Fuzzy_Boolean;
   function "=" (A : Interval_Measure; B : Variable_Measure)
      return Fuzzy_Boolean;
   function "=" (A : Fuzzy_Measure; B : Variable_Measure)
      return Fuzzy_Boolean; 
   function "=" (A : Variable_Measure; B : Measure)
      return Fuzzy_Boolean; 
   function "=" (A : Variable_Measure; B : Interval_Measure)
      return Fuzzy_Boolean; 
   function "=" (A : Variable_Measure; B : Fuzzy_Measure)
      return Fuzzy_Boolean;
   pragma Inline ("=");
--
-- /= -- Inequality
--
--   A - Dimensioned number, interval, fuzzy number, variable
--   B - Dimensioned number, interval, fuzzy number, variable
--
-- At least one argument should be a dimensioned variable.
--
-- Returns :
--
--    not (A = B)
-- 
   function "/=" (A, B : Variable_Measure) return Fuzzy_Boolean; 
   function "/=" (A : Measure; B : Variable_Measure)
      return Fuzzy_Boolean;
   function "/=" (A : Interval_Measure; B : Variable_Measure)
      return Fuzzy_Boolean; 
   function "/=" (A : Fuzzy_Measure; B : Variable_Measure)
      return Fuzzy_Boolean; 
   function "/=" (A : Variable_Measure; B : Measure)
      return Fuzzy_Boolean; 
   function "/=" (A : Variable_Measure; B : Interval_Measure)
      return Fuzzy_Boolean; 
   function "/=" (A : Variable_Measure; B : Fuzzy_Measure)
      return Fuzzy_Boolean;
   pragma Inline ("/=");
--
-- & -- Appending a point to a lingquistic variable
--
--    Left  - The first argument (either a variable or a pair)
--    Right - The second argument (a pair)
--
-- These  functions  are  equivalent  to Append. When Left and Right are
-- pairs  then  both  shall  have exactly same units. Otherwise Left and
-- Right can be differently shifted. 
--
-- Returns :
--
--    The result
--
-- Exceptions :
--
--    Data_Error - The point provided by Right is out of order
--    Unit_Error - Incompatible units
--
   function "&" (Left : Pair_Measure; Right : Pair_Measure)
      return Variable_Measure;
   function "&" (Left : Variable_Measure; Right : Pair_Measure)
      return Variable_Measure;
--
-- and -- Intersection
--
--    A : The first argument
--    B : The second argument
--
-- Returns :
--
--    A and B
--
-- Exceptions :
--
--    Unit_Error - Incompatible units
--
   function "and" (A, B : Variable_Measure) return Variable_Measure;
   function "and" (A : Variable_Measure; B : Confidence)
      return Variable_Measure;
   function "and" (A : Confidence; B : Variable_Measure)
      return Variable_Measure;
   pragma Inline ("and");
--
-- not -- Complement
--
--    A : The argument
--
-- Returns :
--
--    not A
--
   function "not" (A : Variable_Measure) return Variable_Measure;
   pragma Inline ("not");
--
-- or -- Union
--
--    A : The first argument
--    B : The second argument
--
-- Returns :
--
--    A or B
--
-- Exceptions :
--
--    Unit_Error - Incompatible units
--
   function "or" (A, B : Variable_Measure) return Variable_Measure;
   function "or" (A : Variable_Measure; B : Confidence)
      return Variable_Measure;
   function "or" (A : Confidence; B : Variable_Measure)
      return Variable_Measure;
   pragma Inline ("or");
--
-- xor -- Distance
--
--    A : The first argument
--    B : The second argument
--
-- Returns :
--
--    A xor B
--
-- Exceptions :
--
--    Unit_Error - Incompatible units
--
   function "xor" (A, B : Variable_Measure) return Variable_Measure;
   function "xor" (A : Variable_Measure; B : Confidence)
      return Variable_Measure;
   function "xor" (A : Confidence; B : Variable_Measure)
      return Variable_Measure;
   pragma Inline ("xor");

end Fuzzy.Linguistics.Measures;
