--                                                                    --
--  package Fuzzy.Linguistics       Copyright (c)  Dmitry A. Kazakov  --
--  Implementation                                 Luebeck            --
--                                                 Winter, 2003       --
--                                                                    --
--                                Last revision :  14:28 24 Jul 2005  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Ada.IO_Exceptions;  use Ada.IO_Exceptions;

package body Fuzzy.Linguistics is
   use Fuzzy_Floats.Fuzzy_Floats;
   use Points;

   type Operand
        (  Size : Positive;
           Set  : access Points.Set
        )  is limited
   record
      Index : Positive := 1;
      Left  : Confidence;
      This  : Point;
   end record;

   procedure Next (Data : in out Operand) is
   begin
      Data.Left  := Data.This.Right;
      Data.Index := Data.Index + 1;
      if Data.Index <= Data.Size then
         Data.This := Get (Data.Set.all, Data.Index);
      end if;
   end Next;

   function "=" (Left, Right : Point) return Boolean is
   begin
      return Left.Value = Right.Value;
   end "=";

   function "<" (Left, Right : Point) return Boolean is
   begin
      return Left.Value < Right.Value;
   end "<";
--
-- Append_Middle -- Add extreme point on an interval
--
--    Result - A variable to be modified
--    X1     - The left bound of the interval
--    A1, B1 - The values of linear functions there
--    X2     - The right bound of the interval
--    A2, B2 - The values of linear functions there
--
-- This procedure adds a point where A and B intersect. It does  nothing
-- if this point does not belong to ]X1,X2[. 
--
   procedure Append_Middle
             (  Result : in out Variable;
                X1     : Number'Base;
                A1, B1 : Confidence;
                X2     : Number'Base;
                A2, B2 : Confidence
             );
--
-- Append_Xor_Middle -- Add extreme points on an interval
--
--    Result - A variable to be modified
--    X1     - The left bound of the interval
--    A1, B1 - The values of linear functions there
--    X2     - The right bound of the interval
--    A2, B2 - The values of linear functions there
--
-- This procedure adds from one to two points where A and B,  or  A  and
-- not B intersect. It does nothing for a  point  out  of  ]X1,X2[.  The
-- value added in a point is A xor B. 
--
   procedure Append_Xor_Middle
             (  Result : in out Variable;
                X1     : Number'Base;
                A1, B1 : Confidence;
                X2     : Number'Base;
                A2, B2 : Confidence
             );
--
-- Find -- A value in a set of points
--
--    List  - The set
--    Value - The value to search for
--
-- Returns :
--
--    When  positive  it  is  the  index  of the value found in the set.
--    Otherwise it is a negated index of the place the value should be. 
--
   function Find (List : Points.Set; Value : Number)
      return Integer;
   pragma Inline (Find);
--
-- Insert -- New point
--
--    Data - A set of points
--    Item - The new point
--
-- The last point in the set is removed if possible.
--
   procedure Insert
             (  Data : in out Points.Set;
                Item : Point
             );
--
-- Is_In -- Get min-max for a point
--
--    Value - The point
--    Var   - Variable
--    Index - The index of the interval containing Value
--
-- Returns :
--
--    (P(Var|Value), N(Var|Value))
--
   function Is_In (Value : Number; Var : Variable; Index : Integer)
      return Fuzzy_Boolean;

   function Is_Singular (Dot : Point) return Boolean is
   begin
      return Dot.Min = Dot.Max;
   end Is_Singular;

   procedure Insert
             (  Data : in out Points.Set;
                Item : Point
             )  is
      Size : constant Natural := Get_Size (Data);
   begin
      if Size /= 0 then
         declare
            Last : Point renames Get (Data, Size);
         begin
            if (  Is_Singular (Last)
               and then
                  Last.Right = Item.Left
               and then
                  (  Size = 1
                  or else
                     Get (Data, Size - 1).Right = Last.Left
               )  )
            then
               Remove (Data, Size);
            end if;
         end;
      end if;
      Add (Data, Item);
   end Insert;

   function Root
            (  X1     : Number'Base;
               A1, B1 : Confidence;
               X2     : Number'Base;
               A2, B2 : Confidence
            )  return Number'Base is
      DX : constant Number'Base := X2 - X1;
      D1 : constant Number'Base := Number'Base (B1) - Number'Base (A1);
      D2 : constant Number'Base := Number'Base (B2) - Number'Base (A2);
      DD : constant Number'Base := D1 - D2;
   begin
      if abs D1 < abs DD then
         return X1 + D1 * DX / DD;
      else
         return X1;
      end if;
   end Root;

   procedure Append_Middle
             (  Result : in out Variable;
                X1     : Number'Base;
                A1, B1 : Confidence;
                X2     : Number'Base;
                A2, B2 : Confidence
             )  is
       X : constant Number'Base := Root (X1, A1, B1, X2, A2, B2);
   begin
       if X > X1 and then X < X2 then
          Append (Result, X, Interpolate (X, X1, A1, X2, A2));
       end if;
   end Append_Middle;

   procedure Append_Xor_Middle
             (  Result : in out Variable;
                X1     : Number'Base;
                A1, B1 : Confidence;
                X2     : Number'Base;
                A2, B2 : Confidence
             )  is
      procedure Do_X (X : Number'Base; Inversed : Boolean) is
         Y : constant Confidence := Interpolate (X, X1, A1, X2, A2);
      begin
         if X > X1 and then X < X2 then
            if Inversed then
               Append (Result, X, Y or not Y);
            else
               Append (Result, X, Y and not Y);
            end if;
         end if;
      end Do_X;

      X     : constant Number'Base := Root (X1, A1, B1, X2, A2, B2);
      Not_X : constant Number'Base :=
                 Root (X1, A1, not B1, X2, A2, not B2);
   begin
       if X < Not_X then
          Do_X (X,     False);
          Do_X (Not_X, True );
       else
          Do_X (Not_X, True );
          Do_X (X,     False);
       end if;
   end Append_Xor_Middle;

   function Find (List : Points.Set; Value : Number)
      return Integer is
   begin
      return Find
             (  List,
                Point'(Value => Value, others => Confidence'Last)
             );
   end Find;

   function Get_Points_Number (Var : Variable) return Natural is
      Size : Natural := Get_Size (Var.Data);
   begin
      case Size is
         when 0 =>
            return 0;
         when 1 =>
            declare
               This : Point renames Get (Var.Data, 1);
            begin
               if (  This.Min = Confidence'First
                  and then
                     This.Max = Confidence'First
                  )
               then
                  return 0;
               else
                  return 1;
               end if;
            end;
         when others =>
            declare
               This : Point renames Get (Var.Data, Size);
            begin
               if (  Is_Singular (This)
                  and then
                     Get (Var.Data, Size - 1).Right = This.Left
                  )
               then
                  return Size - 1;
               else
                  return Size;
               end if;
            end;
      end case;
   end Get_Points_Number;

   function Interpolate
            (  X  : Number'Base;
               X1 : Number'Base;
               Y1 : Confidence;
               X2 : Number'Base;
               Y2 : Confidence
            )  return Confidence is
   begin
      if Y1 = Y2 then
         return Y1;
      else
         declare
            Y : constant Number'Base :=
               (  Number'Base (Y1)
               +  (  (Number'Base (Y2) - Number'Base (Y1))
                  *  ((X - X1) / (X2 - X1))
               )  );
         begin
            if Y <= Number'Base (Confidence'First) then
               return Confidence'First;
            elsif Y >= Number'Base (Confidence'Last) then
               return Confidence'Last;
            else
               return Confidence (Y);
            end if;
         end;
      end if;
   end Interpolate;

   procedure Append
             (  Var   : in out Variable;
                Value : Number;
                Level : Confidence
             )  is
   begin
      if Get_Size (Var.Data) /= 0 then
         declare
            Last : Point renames Get (Var.Data, Get_Size (Var.Data));
         begin
            if Last.Value > Value then
               raise Data_Error;
            elsif Last.Value = Value then
               declare
                  Updated : Point;
               begin
                  Updated.Value := Last.Value;
                  Updated.Left  := Last.Left;
                  Updated.Right := Level;
                  Updated.Max := Confidence'Max (Last.Max, Level);
                  Updated.Min := Confidence'Min (Last.Min, Level);
                  if not Equal (Updated, Last) then
                     Replace (Var.Data, Updated);
                  end if;
                  return;
               end;
            end if;
         end;
      end if;
      Insert (Var.Data, Point'(Value => Value, others => Level));
   end Append;

   procedure Erase (Var : in out Variable) is
   begin
      Erase (Var.Data);
   end Erase;

   procedure Binary_Operation
             (  Data : in out User_Data;
                A, B : Variable
             )  is separate;
   procedure Unary_Operation
             (  Data : in out User_Data;
                A    : Variable
             )  is separate;

   function Is_In (Value : Number; Var : Variable; Index : Integer)
      return Fuzzy_Boolean is
      Size : constant Natural := Get_Size (Var.Data);
   begin
      if Size = 0 then
         return Certain_False;
      end if;
      if Index > 0 then
         --
         -- There is a point for the argument. The result is  the  value
         -- in  this  point.  Max  gives  the possibility. Min gives the
         -- necessity. 
         --
         declare
            This : Point renames Get (Var.Data, Index);
         begin
            return (Possibility => This.Max, Necessity => This.Min);
         end;
      else
         --
         -- There is no such point we have to interpolate or extrapolate
         -- the membership function. 
         --
         if Index >= -1 then
            --
            -- Extrapolate to the left
            --
            declare
               Result : constant Confidence :=
                  Get (Var.Data, 1).Left;
            begin
               return (Possibility | Necessity => Result);
            end;
         elsif Index = -Size - 1 then
            --
            -- Extrapolate to the right 
            --
            declare
               Result : constant Confidence :=
                  Get (Var.Data, Get_Size (Var.Data)).Right;
            begin
               return (Possibility | Necessity => Result);
            end;
         else
            --
            -- Interpolate between two points
            --
            declare
               Left   : Point renames Get (Var.Data, -Index - 1);
               Right  : Point renames Get (Var.Data, -Index);
               Result : Confidence;
            begin
               Result :=
                  Interpolate
                  (  Value,
                     Left.Value,
                     Left.Right,
                     Right.Value,
                     Right.Left
                  );
               return (Possibility | Necessity => Result);
            end;
         end if;
      end if;
   end Is_In;

   function Is_Interval (Var : Variable) return Boolean is
      Size : constant Natural := Get_Points_Number (Var);
   begin
      if Size = 2 then
         declare
            Left  : Point renames Get (Var.Data, 1);
            Right : Point renames Get (Var.Data, 2);
         begin
            return
            (  Left.Left = Confidence'First
            and then
               Left.Min = Confidence'First
            and then
               Left.Max = Confidence'Last
            and then
               Left.Right = Confidence'Last
            and then
               Right.Left = Confidence'Last
            and then
               Right.Min = Confidence'First
            and then
               Right.Max = Confidence'Last
            and then
               Right.Right = Confidence'First
            );
         end;
      else
         return False;
      end if;
   end Is_Interval;

   function Is_Singleton (Var : Variable) return Boolean is
      Size : constant Natural := Get_Points_Number (Var);
   begin
      if Size = 1 then
         declare
            This : Point renames Get (Var.Data, 1);
         begin
            return
            (  This.Left = Confidence'First
            and then
               This.Min = Confidence'First
            and then
               This.Max = Confidence'Last
            and then
               This.Right = Confidence'First
            );
         end;
      else
         return False;
      end if;
   end Is_Singleton;

   function Is_Empty (Var : Variable) return Boolean is
   begin
      case Get_Size (Var.Data) is
         when 0 =>
            return True;
         when 1 =>
            declare
               Item : Point renames Get (Var.Data, 1);
            begin
               return
               (  Item.Min = Confidence'First
               and then
                  Item.Max = Confidence'First
               );
            end;
         when others =>
            return False;
      end case;
   end Is_Empty;

   function Equal (A, B : Point; Eps : Number'Base := 0.0)
      return Boolean is
   begin
      return
      (  A.Left = B.Left
      and then
         A.Min = B.Min
      and then
         A.Max = B.Max
      and then
         A.Right = B.Right
      and then
         abs (A.Value - B.Value) <= Eps
      );
   end Equal;

   function Equal (A, B : Pair; Eps : Number'Base := 0.0) return Boolean is
   begin
      return A.Level = B.Level and then abs (A.Value - B.Value) <= Eps;
   end Equal;

   function Equal (A : Variable; B : Pair; Eps : Number'Base := 0.0)
      return Boolean is
   begin
      case Get_Size (A.Data) is
         when 0 =>
            return B.Level = Confidence'First;
         when 1 =>
            declare
               Item : Point renames Get (A.Data, 1);
            begin
               return
               (  Item.Left = B.Level
               and then
                  Item.Min = B.Level
               and then
                  Item.Max = B.Level
               and then
                  Item.Right = B.Level
               );
            end;
         when others =>
            return False;
      end case;
   end Equal;

   function Equal (A : Pair; B : Variable; Eps : Number'Base := 0.0)
      return Boolean is
   begin
      return Equal (B, A, Eps);
   end Equal;

   function Equal (A, B : Variable; Eps : Number'Base := 0.0)
      return Boolean is
      Size : constant Natural := Get_Points_Number (A);
   begin
      if Size /= Get_Points_Number (B) then
         return False;
      elsif Size = 1 then
         declare
            L : Point renames Get (A.Data,  1);
            R : Point renames Get (A.Data, 1);
         begin
            return
            (  Equal (L, R, Eps)
            or else
               (  L.Min = L.Max
               and then
                  R.Min = R.Max
               and then
                  L.Min = R.Min
            )  );
         end;
      else
         for Index in 1..Size loop
            if not Equal (Get (A.Data, Index), Get (B.Data, Index), Eps)
            then
               return False;
            end if;
         end loop;
         return True;
      end if;
   end Equal;
--
-- Possibilities -- Implementation of
--
   package Possibilities is
      function P (A : Variable) return Confidence;
      function P (A : Number;      B : Variable) return Confidence;
      function P (A : Interval;    B : Variable) return Confidence;
      function P (A : Fuzzy_Float; B : Variable) return Confidence;
      function P (A : Variable; B : Variable   ) return Confidence;
   end Possibilities;
   package body Possibilities is separate;

   function Possibility (A : Variable) return Confidence
      renames Possibilities.P;
   function Possibility (A : Number; B : Variable) return Confidence
      renames Possibilities.P;
   function Possibility (A : Interval; B : Variable) return Confidence
      renames Possibilities.P;
   function Possibility (A : Fuzzy_Float; B : Variable)
      return Confidence renames Possibilities.P;
   function Possibility (A : Variable; B : Variable) return Confidence
      renames Possibilities.P;
      
   function Possibility (A : Variable; B : Number) return Confidence is
   begin
      return Possibilities.P (B, A);
   end Possibility;

   function Possibility (A : Variable; B : Interval) return Confidence is
   begin
      return Possibilities.P (B, A);
   end Possibility;

   function Possibility (A : Variable; B : Fuzzy_Float)
      return Confidence is
   begin
      return Possibilities.P (B, A);
   end Possibility;
--
-- Necessities -- Implementation of
--
   package Necessities is
      function N (A : Variable) return Confidence;
      function N (A : Number;      B : Variable   ) return Confidence;
      function N (A : Interval;    B : Variable   ) return Confidence;
      function N (A : Fuzzy_Float; B : Variable   ) return Confidence;
      function N (A : Variable;    B : Variable   ) return Confidence;
      function N (A : Variable;    B : Number     ) return Confidence;
      function N (A : Variable;    B : Interval   ) return Confidence;
      function N (A : Variable;    B : Fuzzy_Float) return Confidence;
   end Necessities;
   package body Necessities is separate;

   function Necessity (A : Variable) return Confidence
      renames Necessities.N;
   function Necessity (A : Number; B : Variable) return Confidence
      renames Necessities.N;
   function Necessity (A : Interval; B : Variable) return Confidence
      renames Necessities.N;
   function Necessity (A : Fuzzy_Float; B : Variable) return Confidence
      renames Necessities.N;
   function Necessity (A : Variable; B : Variable) return Confidence
      renames Necessities.N;
   function Necessity (A : Variable; B : Number) return Confidence
      renames Necessities.N;
   function Necessity (A : Variable; B : Interval) return Confidence
      renames Necessities.N;
   function Necessity (A : Variable; B : Fuzzy_Float) return Confidence
      renames Necessities.N;

   function Is_In (A : Number; B : Variable)
      return Fuzzy_Boolean is
   begin
      return Is_In (A, B, Find (B.Data, A));
   end Is_In;

   function Is_In (A : Interval; B : Variable)
      return Fuzzy_Boolean is
   begin
      if A.From = A.To then
         --
         -- Only one point in the interval
         --
         return Is_In (A.From, B, Find (B.Data, A.From));
      end if;
      --
      -- Several points in the interval
      --
      declare
         From   : Integer := Find (B.Data, A.From);
         To     : Integer := Find (B.Data, A.To);
         Result : Fuzzy_Boolean :=
                     Is_In (A.From, B, From) + Is_In (A.To, B, To);
      begin
         --
         -- Result now contains data obtained from the interval ends. To
         -- complete  it,  we  have  to  go  through  all  points in the
         -- interval. 
         --
         if From < 0 then
            From := -From;
         else
            From := From + 1;
         end if;
         if To <= 0 then
            To := - To - 1;
         else
            To := To - 1;
         end if;
         for Index in From..To loop
            declare
               This : Point renames Get (B.Data, Index);
            begin
               Result.Possibility := Result.Possibility or This.Max;
               Result.Necessity   := Result.Necessity  and This.Min;
            end;
         end loop;
         return Result;
      end;
   end Is_In;

   function Is_In (A : Fuzzy_Float; B : Variable)
      return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (B, A),
         Necessity   => Necessity   (B, A)
      );
   end Is_In;

   function Is_In (A, B : Variable)
      return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (B, A),
         Necessity   => Necessity   (B, A)
      );
   end Is_In;

   function Is_In (A : Variable; B : Number)
      return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (B, A),
         Necessity   => Necessity   (B, A)
      );
   end Is_In;

   function Is_In (A : Variable; B : Interval)
      return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (B, A),
         Necessity   => Necessity   (B, A)
      );
   end Is_In;

   function Is_In (A : Variable; B : Fuzzy_Float)
      return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (B, A),
         Necessity   => Necessity   (B, A)
      );
   end Is_In;

   function "+" (Left : Variable; Right : Number) return Variable is
      Result : Variable;
      This   : Point;
   begin
      for Index in 1..Get_Size (Left.Data) loop
         This := Get (Left.Data, Index);
         This.Value := This.Value + Right;
         Add (Result.Data, This);
      end loop;
      return Result;
   end "+";

   function "-" (Left : Variable; Right : Number) return Variable is
      Result : Variable;
      This   : Point;
   begin
      for Index in 1..Get_Size (Left.Data) loop
         This := Get (Left.Data, Index);
         This.Value := This.Value - Right;
         Add (Result.Data, This);
      end loop;
      return Result;
   end "-";

   function "*" (Left : Variable; Right : Number) return Variable is
      Result : Variable;
      This   : Point;
   begin
      if Right > 0.0 then
         for Index in 1..Get_Size (Left.Data) loop
            This := Get (Left.Data, Index);
            This.Value := This.Value * Right;
            Add (Result.Data, This);
         end loop;
      else
         for Index in reverse 1..Get_Size (Left.Data) loop
            This := Get (Left.Data, Index);
            This.Value := This.Value * Right;
            Add (Result.Data, This);
         end loop;
      end if;
      return Result;
   end "*";

   function "/" (Left : Variable; Right : Number) return Variable is
      Result : Variable;
      This   : Point;
   begin
      if Right > 0.0 then
         for Index in 1..Get_Size (Left.Data) loop
            This := Get (Left.Data, Index);
            This.Value := This.Value / Right;
            Add (Result.Data, This);
         end loop;
      else
         for Index in reverse 1..Get_Size (Left.Data) loop
            This := Get (Left.Data, Index);
            This.Value := This.Value / Right;
            Add (Result.Data, This);
         end loop;
      end if;
      return Result;
   end "/";

   function "=" (A : Number; B : Variable) return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (A, B),
         Necessity   => Necessity (A, B) and Necessity (B, A)
      );
   end "=";

   function "=" (A : Interval; B : Variable) return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (A, B),
         Necessity   => Necessity (A, B) and Necessity (B, A)
      );
   end "=";

   function "=" (A : Fuzzy_Float; B : Variable) return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (A, B),
         Necessity   => Necessity (A, B) and Necessity (B, A)
      );
   end "=";

   function "=" (A : Variable; B : Variable) return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (A, B),
         Necessity   => Necessity (A, B) and Necessity (B, A)
      );
   end "=";
 
   function "=" (A : Variable; B : Number) return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (A, B),
         Necessity   => Necessity (A, B) and Necessity (B, A)
      );
   end "=";
 
   function "=" (A : Variable; B : Interval) return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (A, B),
         Necessity   => Necessity (A, B) and Necessity (B, A)
      );
   end "=";
 
   function "=" (A : Variable; B : Fuzzy_Float) return Fuzzy_Boolean is
   begin
      return
      (  Possibility => Possibility (A, B),
         Necessity   => Necessity (A, B) and Necessity (B, A)
      );
   end "=";

   function "/=" (A : Number; B : Variable) return Fuzzy_Boolean is
   begin
      return not (A = B);
   end "/=";

   function "/=" (A : Interval; B : Variable) return Fuzzy_Boolean is
   begin
      return not (A = B);
   end "/=";

   function "/=" (A : Fuzzy_Float; B : Variable) return Fuzzy_Boolean is
   begin
      return not (A = B);
   end "/=";

   function "/=" (A : Variable; B : Variable) return Fuzzy_Boolean is
   begin
      return not (A = B);
   end "/=";
 
   function "/=" (A : Variable; B : Number) return Fuzzy_Boolean is
   begin
      return not (A = B);
   end "/=";
 
   function "/=" (A : Variable; B : Interval) return Fuzzy_Boolean is
   begin
      return not (A = B);
   end "/=";
 
   function "/=" (A : Variable; B : Fuzzy_Float) return Fuzzy_Boolean is
   begin
      return not (A = B);
   end "/=";

   function "&" (Left : Pair; Right : Pair) return Variable is
      Result : Variable;
   begin
      Append (Result, Left.Value,  Left.Level);
      Append (Result, Right.Value, Right.Level);
      return Result;
   end "&";

   function "&" (Left : Variable; Right : Pair) return Variable is
      Result : Variable := Left;
   begin
      Append (Result, Right.Value, Right.Level);
      return Result;
   end "&";
--
-- Operations -- Implementation of
--
--    **            - The inducing dot operation
--    Append_Middle - To deal with an interval
--
   generic
      with function "**" (A, B : Confidence) return Confidence;
      with procedure Append_Middle
                     (  Result : in out Variable;
                        X1     : Number'Base;
                        A1, B1 : Confidence;
                        X2     : Number'Base;
                        A2, B2 : Confidence
                     )  is <>;
   package Operations is
      function Operation (A, B : Variable) return Variable;
      function Operation (A : Variable; B : Confidence) return Variable;
   end Operations;
   package body Operations is separate;

   package And_Operations is new Operations ("and");
   function "and" (A, B : Variable) return Variable
      renames And_Operations.Operation;
   function "and" (A : Variable; B : Confidence) return Variable
      renames And_Operations.Operation;
   function "and" (A : Confidence; B : Variable) return Variable is
   begin
      return B and A;
   end "and";

   package Or_Operations is new Operations ("or");
   function "or" (A, B : Variable) return Variable
      renames Or_Operations.Operation;
   function "or" (A : Variable; B : Confidence) return Variable
      renames Or_Operations.Operation;
   function "or" (A : Confidence; B : Variable) return Variable is
   begin
      return B and A;
   end "or";

   package Xor_Operations is new Operations ( "xor", Append_Xor_Middle);
   function "xor" (A, B : Variable) return Variable
      renames Xor_Operations.Operation;
   function "xor" (A : Variable; B : Confidence) return Variable
      renames Xor_Operations.Operation;
   function "xor" (A : Confidence; B : Variable) return Variable is
   begin
      return B and A;
   end "xor";

   function "not" (A : Variable) return Variable is
      Size   : constant Natural := Get_Points_Number (A);
      Result : Variable;
   begin
      if Size = 0 then
         Append (Result, Number'First, Confidence'Last);
      else
         declare
            This : Point;
         begin
            for Index in 1..Size loop
               This := Get (A.Data, Index);
               Append (Result, This.Value, not This.Left);
               if This.Left /= This.Min then
                  Append (Result, This.Value, not This.Min);
               end if;
               if This.Max /= This.Min then
                  Append (Result, This.Value, not This.Max);
               end if;
               if This.Right /= This.Max then
                  Append (Result, This.Value, not This.Right);
               end if;
            end loop;
         end;
      end if;
      return Result;
   end "not";

   function Empty return Variable is
      Result : Variable;
   begin
      return Result;
   end Empty;

   function To_Interval (Value : Variable) return Interval is
   begin
      if Is_Interval (Value) then
         return (Get (Value.Data, 1).Value, Get (Value.Data, 2).Value);
      else
         raise Constraint_Error;
      end if;
   end To_Interval;

   function To_Number (Value : Variable) return Number is
   begin
      if Is_Interval (Value) then
         return Get (Value.Data, 1).Value;
      else
         raise Constraint_Error;
      end if;
   end To_Number;

   function To_Variable (Value : Number) return Variable is
      Result : Variable;
   begin
      Append (Result, Value, Confidence'First);
      Append (Result, Value, Confidence'Last);
      Append (Result, Value, Confidence'First);
      return Result;
   end To_Variable;

   function To_Variable (Value : Interval) return Variable is
      Result : Variable;
   begin
      Append (Result, Value.From, Confidence'First);
      Append (Result, Value.From, Confidence'Last);
      Append (Result, Value.To,   Confidence'Last);
      Append (Result, Value.To,   Confidence'First);
      return Result;
   end To_Variable;

   function To_Variable (Value : Fuzzy_Float) return Variable is
      Result : Variable;
      X      : Number;
   begin
      X := Get_Interval (Value, Interval_Index'First).From;
      Append (Result, X, Confidence'First);
      Append
      (  Result,
         X,
         Fuzzy_Floats.To_Confidence (Interval_Index'First)
      );
      for Index in Interval_Index'Succ (Interval_Index'First)
                .. Interval_Index'Last
      loop
         X := Get_Interval (Value, Index).From;
         Append (Result, X, Fuzzy_Floats.To_Confidence (Index));
      end loop;
      for Index in reverse Interval_Index'First..Interval_Index'Last
      loop
         X := Get_Interval (Value, Index).To;
         Append (Result, X, Fuzzy_Floats.To_Confidence (Index));
      end loop;
      Append (Result, X, Confidence'First);
      return Result;
   end To_Variable;

end Fuzzy.Linguistics;
