--                                                                    --
--  package Fuzzy.Linguistics       Copyright (c)  Dmitry A. Kazakov  --
--  Interface                                      Luebeck            --
--                                                 Winter, 2003       --
--                                                                    --
--                                Last revision :  15:51 07 May 2005  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
--  This generic  package  provides  linguistic  variables.  The  formal
--  generic parameter is an package implementing fuzzy floats,  i.e.  an
--  instantiation of Fuzzy.Floats: 
--
--     Fuzzy_Floats - An instance of Fuzzy.Floats
--
--  A  linguistic  variable  is  a  fuzzy  subset  of  real numbers. The
--  membership function of a  linguistic  variable  consists  of  linear
--  segments and singletons. It is defined by a finite set of points. In
--  each point there are values defined: 
--
--     (o)  left limit
--     (o)  right limit
--     (o)  possibility at the point
--     (o)  necessity at the point
--
--  This allows to represent  functions built as a union of  trapezoids,
--  rectangles and singular points. 
--
with Fuzzy.Floats;
with Fuzzy.Logic;     use Fuzzy.Logic;
with Generic_Set;

generic
   with package Fuzzy_Floats is new Fuzzy.Floats (<>);
package Fuzzy.Linguistics is
   use Fuzzy_Floats;
   use Float_Intervals;
--
-- Variable -- A linguistic variable
--
   type Variable is private;
--
-- Pair -- A value - confidence level pair
--
   type Pair is record
      Value : Number;
      Level : Confidence;
   end record;
--
-- Append -- One point to a linguistic variable
--
--    Var   - The variable
--    Value - The membership function argument
--    Level - The membership function value in the argument
--
-- This function is used to form the membership function of a linguistic
-- variable. It adds a new point to the  function.  Points  have  to  be
-- added in ascending order. It is allowed to define several points with
-- same argument (Value). 
--
-- Exceptions :
--
--    Data_Error - The point is out of order
--
   procedure Append
             (  Var   : in out Variable;
                Value : Number;
                Level : Confidence
             );
--
-- Empty -- Empty variable
--
   function Empty return Variable;
   pragma Inline (Empty);
--
-- Erase -- Makes a variable an empty set
--
--    Var - The variable
--
   procedure Erase (Var : in out Variable);
   pragma Inline (Erase);
--
-- Equal -- Compare two variables (equality)
--
--    A   - The first argument
--    B   - The second argument
--    Eps - Tolerance level
--
-- One of the arguments can be a value-confidence pair. Two  points  are
-- assumed same if their confidence levels are same and values differ no
-- more than Eps.
--
-- Returns :
--
--    True if A and B are exactly same
--
   function Equal (A, B : Variable; Eps : Number'Base := 0.0)
      return Boolean;
   function Equal (A : Variable; B : Pair; Eps : Number'Base := 0.0)
      return Boolean;
   function Equal (A : Pair; B : Variable; Eps : Number'Base := 0.0)
      return Boolean;
   function Equal (A, B : Pair; Eps : Number'Base := 0.0) return Boolean;
--
-- Get_Points_Number -- Get number of points of the membership function
--
--    Var - The variable
--
-- Returns :
--
--    The number of distinct points of the membership function
--
   function Get_Points_Number (Var : Variable) return Natural;
   pragma Inline (Get_Points_Number);
--
-- Is_Empty -- Check if a variable is an empty set
--
--    Var - The variable
--
-- Returns :
--
--    True if the membership function is everywhere False
--
   function Is_Empty (Var : Variable) return Boolean;
--
-- Is_In -- Membership / subset test
--
--    A - Number, interval, fuzzy number, a lingustic variable
--    B - Number, interval, fuzzy number, a lingustic variable
--
-- At least one argument should be a Variable.
--
-- Returns :
--
--    (P(B|A), N(B|A))
--
   function Is_In (A, B : Variable) return Fuzzy_Boolean;
   function Is_In (A : Number;      B : Variable) return Fuzzy_Boolean;
   function Is_In (A : Interval;    B : Variable) return Fuzzy_Boolean; 
   function Is_In (A : Fuzzy_Float; B : Variable) return Fuzzy_Boolean;
   function Is_In (A : Variable; B : Number)      return Fuzzy_Boolean;
   function Is_In (A : Variable; B : Interval)    return Fuzzy_Boolean;
   function Is_In (A : Variable; B : Fuzzy_Float) return Fuzzy_Boolean;
--
-- Is_Interval -- Check if a variable is an interval
--
--    Var - The variable
--
-- This function returns True if the membership function of Var is 1  on
-- a finite interval and 0 outside it.
--
-- Returns :
--
--    True if it is an interval
--
   function Is_Interval (Var : Variable)  return Boolean;
--
-- Is_Singleton -- Check if a variable is an singleton
--
--    Var   - The variable
--
-- This function returns True if the membership function of Vae is 1 in
-- one point and 0 elsewhere.
--
-- Returns :
--
--    True if is a singleton
--
   function Is_Singleton (Var : Variable) return Boolean;
--
-- Necessity -- Conditional necessity
--
--    A - The first argument
--    B - The second argument (condition)
--
-- With  one  argument the condition is assumed to be the domain set. At
-- least one argument should be a Variable.
--
-- Returns :
--
--    N(A|B)
--
   function Necessity (A : Variable) return Confidence;
   function Necessity (A, B : Variable) return Confidence;
   function Necessity (A : Number;      B : Variable) return Confidence;
   function Necessity (A : Interval;    B : Variable) return Confidence;
   function Necessity (A : Fuzzy_Float; B : Variable) return Confidence;
   function Necessity (A : Variable; B : Number     ) return Confidence;
   function Necessity (A : Variable; B : Interval   ) return Confidence;
   function Necessity (A : Variable; B : Fuzzy_Float) return Confidence;
--
-- Possibility -- Conditional possibility
--
--    A - The first argument
--    B - The second argument (condition)
--
-- With  one  argument the condition is assumed to be the domain set. At
-- least one argument should be a Variable. 
--
-- Returns :
--
--    P(A|B)
--
   function Possibility (A : Variable) return Confidence;
   function Possibility (A, B : Variable) return Confidence;
   function Possibility (A : Number;      B : Variable) return Confidence;
   function Possibility (A : Interval;    B : Variable) return Confidence;
   function Possibility (A : Fuzzy_Float; B : Variable)
      return Confidence;
   function Possibility (A : Variable; B : Number     ) return Confidence;
   function Possibility (A : Variable; B : Interval   ) return Confidence;
   function Possibility (A : Variable; B : Fuzzy_Float)
      return Confidence;
--
-- To_Interval -- Conversion
--
--    Value - A variable to convert
--
-- Returns :
--
--    The corresponding interval
--
-- Exceptions :
--
--    Constraint_Error - Value is not an interval
--
   function To_Interval (Value : Variable) return Interval;
--
-- To_Number -- Conversion
--
--    Value - A variable to convert
--
-- Returns :
--
--    The corresponding number
--
-- Exceptions :
--
--    Constraint_Error - Value is not a singleton (number)
--
   function To_Number (Value : Variable) return Number;
--
-- To_Variable -- Conversion
--
--    Value - A number, interval or a fuzzy number
--
-- Returns :
--
--    The corresponding variable
--
   function To_Variable (Value : Number) return Variable;
   function To_Variable (Value : Interval) return Variable;
   function To_Variable (Value : Fuzzy_Float) return Variable;
--
-- +, -, *, / -- By a number
--
--    Left  - Variable
--    Right - Number
--
-- Returns :
--
--    The result of the operation
--
   function "+" (Left : Variable; Right : Number) return Variable;
   function "-" (Left : Variable; Right : Number) return Variable;
   function "*" (Left : Variable; Right : Number) return Variable;
   function "/" (Left : Variable; Right : Number) return Variable;
--
-- = -- Equality
--
--    A - Number, Interval, Fuzzy number, Variable
--    B - Number, Interval, Fuzzy number, Variable
--
-- At least one argument should be a Variable.
--
-- Returns :
--
--    (P(A|B), N(A|B) and N(B|A))
--
   function "=" (A, B : Variable) return Fuzzy_Boolean; 
   function "=" (A : Number;      B : Variable) return Fuzzy_Boolean;
   function "=" (A : Interval;    B : Variable) return Fuzzy_Boolean; 
   function "=" (A : Fuzzy_Float; B : Variable) return Fuzzy_Boolean; 
   function "=" (A : Variable; B : Number)      return Fuzzy_Boolean; 
   function "=" (A : Variable; B : Interval)    return Fuzzy_Boolean; 
   function "=" (A : Variable; B : Fuzzy_Float) return Fuzzy_Boolean;
   pragma Inline ("=");
--
-- /= -- Inequality
--
--    A - Number, Interval, Fuzzy number, Variable
--    B - Number, Interval, Fuzzy number, Variable
--
-- At least one argument should be a Variable.
--
-- Returns :
--
--    not (A = B)
-- 
   function "/=" (A, B : Variable) return Fuzzy_Boolean; 
   function "/=" (A : Number;      B : Variable) return Fuzzy_Boolean;
   function "/=" (A : Interval;    B : Variable) return Fuzzy_Boolean; 
   function "/=" (A : Fuzzy_Float; B : Variable) return Fuzzy_Boolean; 
   function "/=" (A : Variable; B : Number)      return Fuzzy_Boolean; 
   function "/=" (A : Variable; B : Interval)    return Fuzzy_Boolean; 
   function "/=" (A : Variable; B : Fuzzy_Float) return Fuzzy_Boolean;
   pragma Inline ("/=");
--
-- & -- Appending a point to a lingquistic variable
--
--    Left  - The first argument (either a variable or a pair)
--    Right - The second argument (a pair)
--
-- These functions are equivalent to Append.
--
-- Returns :
--
--    The result
--
-- Exceptions :
--
--    Data_Error - The point provided by Right is out of order
--
   function "&" (Left : Pair;     Right : Pair) return Variable;
   function "&" (Left : Variable; Right : Pair) return Variable;
--
-- and -- Intersection
--
--    A : The first argument
--    B : The second argument
--
-- Returns :
--
--    A and B
--
   function "and" (A, B : Variable) return Variable;
   function "and" (A : Variable;   B : Confidence) return Variable;
   function "and" (A : Confidence; B : Variable)   return Variable;
--
-- not -- Complement
--
--    A : The argument
--
-- Returns :
--
--    not A
--
   function "not" (A : Variable) return Variable;
--
-- or -- Union
--
--    A : The first argument
--    B : The second argument
--
-- Returns :
--
--    A or B
--
   function "or" (A, B : Variable) return Variable;
   function "or" (A : Variable;   B : Confidence) return Variable;
   function "or" (A : Confidence; B : Variable)   return Variable;
--
-- xor -- Distance
--
--    A : The first argument
--    B : The second argument
--
-- Returns :
--
--    A xor B
--
   function "xor" (A, B : Variable) return Variable;
   function "xor" (A : Variable;   B : Confidence) return Variable;
   function "xor" (A : Confidence; B : Variable)   return Variable;

private
--
-- Point -- One interpolation point of a membership function
--
--    Value - The argument of the function
--    Left  - The left limit of the function
--    Right - The right limit of the function
--    Max   - The upper value of the function in the argument
--    Min   - The lower value of the function in the argument
--
   type Point is record
      Value : Number;
      Left  : Confidence := Confidence'First;
      Right : Confidence := Confidence'First;
      Min   : Confidence := Confidence'First;
      Max   : Confidence := Confidence'First;
   end record;
--
-- Is_Singular -- Check if a point is a singular value
--
--    Dot : To be tested
--
-- Returns :
--
--    True if all levels are same
--
   function Is_Singular (Dot : Point) return Boolean;
   pragma Inline (Is_Singular);
--
-- Equal -- Compare to points
--
--    A   - The first argument
--    B   - The second argument
--    Eps - Tolerance
--
-- Returns :
--
--    True if Left is equal to Right
--
   function Equal (A, B : Point; Eps : Number'Base := 0.0)
      return Boolean;
   function "=" (Left, Right : Point) return Boolean;
   function "<" (Left, Right : Point) return Boolean;
   pragma Inline (Equal, "<", "=");
   type Point_Array is array (Positive range <>) of Point;
--
-- Points -- Instantiation of generic set
--
-- The package  provides  sets  of  interpolation  points.  Such  a  set
-- represents a membership function.
--   
   package Points is new
      Generic_Set
      (  Object_Type  => Point,
         Null_Element =>
            (  Number'First,
               Confidence'First,
               Confidence'First,
               Confidence'Last,
               Confidence'First
            ),
         Minimal_Size => 8
      );
   use Points;
   type Variable is record
      Data : Points.Set;
   end record;
--
-- Binary_Operation -- Operations on two variables
--
--    User_Data   - The user-defined data type (operation context)
--    Do_Empty    - Action when one of the arguments is an empty set
--    Do_Point    - Action for a singular point
--    Do_Interval - Action on an interval
--
-- An instance is a procedure with the following parameters:
--
--    Data - The user data
--    A    - The first parameter (variable)
--    B    - The first parameter (variable)
--
-- If  either  A  or  B  is  empty, then Do_Empty is called with another
-- parameter. Otherwise, the procedure enumerates the joining points  of
-- A and B. The points are enumerated in ascending order. For each  such
-- point Do_Point is called at least once. The  parameters  of  Do_Point
-- are the values of the membership functions in the point.  The  values
-- are  interpolated  if  necessary. Do_Point is called several times if
-- Left,  Min,  Max,  Right  differs  there. It is warrantied that it is
-- called for Min-Min and Max-Max combinations if they differ. For  each
-- pair of points Do_Interval is called. 
--
   generic
      type User_Data is limited private;
      with procedure Do_Empty (Data : in out User_Data; A : Variable);
      with procedure Do_Point
                     (  Data : in out User_Data;
                        X    : Number'Base;
                        A, B : Confidence
                     );
      with procedure Do_Interval
                     (  Data   : in out User_Data;
                        X1     : Number'Base;
                        A1, B1 : Confidence;
                        X2     : Number'Base;
                        A2, B2 : Confidence
                     );
   procedure Binary_Operation (Data : in out User_Data; A, B : Variable);
--
-- Unary_Operation -- Operations on a variable
--
--    User_Data   - The user-defined data type (operation context)
--    Do_Empty    - Action when one of the arguments is an empty set
--    Do_Point    - Action for a singular point
--    Do_Interval - Action on an interval
--
-- An instance is a procedure with the following parameters:
--
--    Data - The user data
--    A    - The parameter (variable)
--
-- If A is empty, then Do_Empty  is  called.  Otherwise,  the  procedure
-- enumerates the joining points of A.  The  points  are  enumerated  in
-- ascending order. For each such point  Do_Point  is  called  at  least
-- once.  Do_Point  is  called  several  times  if Left, Min, Max, Right
-- differs there. For each pair of points Do_Interval is called. 
--
   generic
      type User_Data is limited private;
      with procedure Do_Empty (Data : in out User_Data);
      with procedure Do_Point
                     (  Data : in out User_Data;
                        X    : Number'Base;
                        Y    : Confidence
                     );
      with procedure Do_Interval
                     (  Data : in out User_Data;
                        X1   : Number'Base;
                        Y1   : Confidence;
                        X2   : Number'Base;
                        Y2   : Confidence
                     );
   procedure Unary_Operation (Data : in out User_Data; A : Variable);
--
-- Interpolate -- Confidence factors (linear)
--
--    X  - The argument
--    X1 - The first interpolation point
--    Y1 - The value at this point
--    X2 - The second interpolation point
--    Y2 - The value there
--
-- Returns :
--
--    An interpolated value in X
--
   function Interpolate
            (  X  : Number;
               X1 : Number;
               Y1 : Confidence;
               X2 : Number;
               Y2 : Confidence
            )  return Confidence;
   pragma Inline (Interpolate);
--
-- Root -- Intersection of two linear functions on an interval
--
--    X1     - The left bound of the interval
--    A1, B1 - The function values there
--    X2     - The right bound
--    A2, B2 - The function values there
--
-- The  result  is  valid if in ]X1,X2[. Otherwise, the functions do not
-- intersect on ]X1, X2[ and the result has to be ignored.
--
-- Returns :
--
--    The result
--
   function Root
            (  X1     : Number;
               A1, B1 : Confidence;
               X2     : Number;
               A2, B2 : Confidence
            )  return Number'Base;
end Fuzzy.Linguistics;
