--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Fuzzy.Measures                              Luebeck            --
--  Interface                                      Spring, 2005       --
--                                                                    --
--                                Last revision :  20:52 06 Jun 2005  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
--  This  package provides dimensioned fuzzy floating-point numbers. The
--  package is generic, the formal parameter are: 
--
--     Fuzzy_Floats      - Floating-point fuzzy numbers
--     Float_Measures    - Compatible dimensioned numbers
--     Interval_Measures - Compatible dimensioned intervals
--
--  The behavior of the operations on  fuzzy  numbers  on  overflows  is
--  determined  by  one  of  the  underlying  floating-point  type.  See
--  T'Machine_Overflows. 
--
with Fuzzy.Logic;  use Fuzzy.Logic;
with Units;        use Units;

with Fuzzy.Floats;
with Intervals.Measures;
with Measures;
with Units.Base;

generic
   with package Fuzzy_Floats is new Fuzzy.Floats (<>);
   with package Float_Measures is
      new Standard.Measures (Fuzzy_Floats.Number);
   with package Interval_Measures is
      new Intervals.Measures
          (  Float_Measures,
             Fuzzy_Floats.Float_Intervals
          );
package Fuzzy.Measures is
   use Float_Measures;
   use Fuzzy_Floats;
   use Fuzzy_Floats.Float_Intervals;
   use Interval_Measures;

   subtype Number is Fuzzy_Floats.Number;
--
-- Fuzzy_Measure -- Dimensioned fuzzy number
--
--    SI     - The dimension unit
--    Gain   - The value of, in the dimension units
--    Offset - The dimension unit shift
--
   type Fuzzy_Measure (SI : Unit := Units.Base.Unitless) is record
      Gain   : Fuzzy_Float;
      Offset : Number'Base := 0.0;
   end record;
--
-- Convert -- Measure conversion
--
--    Value - The fuzzy measurement to convert
--    Scale - The measure of the result
--
-- This function is used  to  convert  a  fuzzy  measure  Value  to  the
-- measurement units specified by the parameter Scale. When  offsets  of
-- Value and Scale are same this is the null operation. 
--
--    Convert (T, Celsius)  -- Temperature in Celsius degrees
--
-- Returns :
--
--    Scale equivalent of Value
--
-- Exceptions :
--
--    Unit_Error -- Item and Scale have different units
--
   function Convert (Value : Fuzzy_Measure; Scale : Measure)
      return Fuzzy_Measure;
   pragma Inline (Convert);   
--
-- Equal -- Equality
--
--    Left  - The first argument
--    Right - The first argument
--
-- Returns :
--
--    True if Left is identic to Right
--
   function Equal (Left, Right : Fuzzy_Measure) return Boolean;
--
-- Get_Value -- Get SI value
--
--    Value  - The fuzzy measure
--
-- Returns :
--
--    SI equivalent of Value as a fuzzy number
--
   function Get_Value (Value : Fuzzy_Measure) return Fuzzy_Float;
   pragma Inline (Get_Value);
--
-- Get_Value_As -- Get value measured in non-SI units
--
--    Value  - The fuzzy measure
--    Scale  - The measure of the result
--
-- This  function  is  used to get the value in units other than SI. For
-- instance: 
--
--    Get_Value_As (T, Celsius)  -- Temperatures in Celsius degrees
--
-- Returns :
--
--    Scale equivalent of Value in as a fuzzy number
--
-- Exceptions :
--
--    Unit_Error -- Value and Scale have different units
--
   function Get_Value_As (Value : Fuzzy_Measure; Scale : Measure)
      return Fuzzy_Float;
   pragma Inline (Get_Value_As);
--
-- Get_Unit -- Get unit
--
--    Value  - The measure 
--
-- Returns :
--
--    SI component
--
   function Get_Unit (Value : Fuzzy_Measure) return Unit;
   pragma Inline (Get_Unit);   
--
-- Is_In -- Membership test
--
--    Left  - The first argument
--    Right - The second argument
--
-- One of the parameters shall be a fuzzy measure. Other can be interval
-- measure, measure or fuzzy measure. Tests can be applied to parameters
-- of different units, in which case the result is Certain_False. 
--
-- Returns :
--
--    ( Possibility (Right | Left), Necessity (Right | Left) )
--
   function Is_In (Left, Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function Is_In (Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function Is_In (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function Is_In (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Boolean;
   function Is_In (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Boolean;
--
-- Necessity -- The possibility theory
--
--    Left  - A measure / interval measure / fuzzy measure
--    Right - A measure / interval measure / fuzzy measure
--
-- The result tells how  necessary  is  Left  given  Right.  It  is  the
-- inversed confidence level of the most nested interval  of  Right  not
-- intersecting  Left.  The  arguments  may  have  different shifts. The
-- result is a lower estimation of the precise result. 
--
-- Returns :
--
--    Necessity (Left | Right}
--
-- Exceptions :
--
--    Unit_Error - Incompatible units
--
   function Necessity (Left, Right : Fuzzy_Measure)
      return Confidence;
   function Necessity (Left : Measure; Right : Fuzzy_Measure)
      return Confidence;
   function Necessity (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Confidence;
   function Necessity (Left : Fuzzy_Measure; Right : Measure)
      return Confidence;
   function Necessity (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Confidence;
--
-- Normalize -- Shift removing
--
--    Value - Fuzzy measure
--
-- This  function  is  used  to  convert  Value  to  to  its   unshifted
-- equivalent. 
--
-- Returns :
--
--    Unshifted equivalent of Value
--
   function Normalize (Value : Fuzzy_Measure) return Fuzzy_Measure;
   pragma Inline (Normalize);
--
-- Possibility -- The possiblity theory
--
--    Left  - A measure / interval measure / fuzzy measure
--    Right - A measure / interval measure / fuzzy measure
--
-- The  result  tells  how  possible  is  Left  given  Right.  It is the
-- confidence  level  of  the most nested interval of Right intersecting
-- Left. The arguments may have different shifts.
--
-- Returns :
--
--    Possibility (Left | Right}
--
-- Exceptions :
--
--    Unit_Error - Incompatible units
--
   function Possibility (Left, Right : Fuzzy_Measure)
      return Confidence;
   function Possibility (Left : Measure; Right : Fuzzy_Measure)
      return Confidence;
   function Possibility (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Confidence;
   function Possibility (Left : Fuzzy_Measure; Right : Measure)
      return Confidence;
   function Possibility (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Confidence;
--
-- Shift -- Non-destructive shift
--
--    Value  - Fuzzy measure
--    Shift  - Shift
--
-- This function is used to convert Value to to its shifted equivalent. 
--
-- Returns :
--
--    Shifted equivalent of Value
--
   function Shift (Value : Fuzzy_Measure; Shift : Number'Base)
      return Fuzzy_Measure;
   pragma Inline (Shift);
--
-- To_Fuzzy -- Conversion to fuzzy
--
--    Left - Number, measure, interval, interval measure, fuzzy number
--
-- Returns :
--
--    Fuzzy equivalent
--
   function To_Fuzzy_Measure (Left : Number)
      return Fuzzy_Measure;
   function To_Fuzzy_Measure (Left : Measure)
      return Fuzzy_Measure;
   function To_Fuzzy_Measure (Left : Interval)
      return Fuzzy_Measure;
   function To_Fuzzy_Measure (Left : Interval_Measure)
      return Fuzzy_Measure;
   function To_Fuzzy_Measure (Left : Fuzzy_Float)
      return Fuzzy_Measure;
   pragma Inline (To_Fuzzy_Measure);
--
-- abs -- Absolute value
--
--    Left - The argument
--
-- Returns :
--
--    Absolute value
--
   function "abs" (Left : Fuzzy_Measure) return Fuzzy_Measure;
--
-- + -- Unary plus
--
--    Left - The operand
--
-- Returns :
--
--    Left
--
   function "+" (Left : Fuzzy_Measure) return Fuzzy_Measure;
--
-- - -- Unary minus
--
--    Left - The operand
--
-- Returns :
--
--    -Left
--
   function "-" (Left : Fuzzy_Measure) return Fuzzy_Measure;
--
-- + -- Addition
--
--    Left  - A measure / interval measure / fuzzy measure
--    Right - A measure / interval measure / fuzzy measure
--
-- Returns :
--
--      Left + Right
--
-- Exceptions :
--
--    Unit_Error - Incompatible units
--
   function "+" (Left, Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "+" (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Measure;
   function "+" (Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "+" (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Measure;
   function "+" (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
--
-- - -- Subtraction
--
--    Left  - A measure / interval measure / fuzzy measure
--    Right - A measure / interval measure / fuzzy measure
--
-- Returns :
--
--      Left - Right
--
-- Exceptions :
--
--    Unit_Error - Incompatible units
--
   function "-" (Left, Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "-" (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Measure;
   function "-" (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "-" (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Measure;
   function "-" (Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
--
-- * -- Multiplication
--
--    Left  - First argument
--    Right - Second argument
--
-- One  of  the  argument  is  a  fuzzy measure another is either of the
-- following types:  fuzzy  measure,  fuzzy  number,  number,  interval,
-- interval  measure.  Additionally  multiplication  is  defined  on   a
-- dimensioned number or interval and a  fuzzy  number  resulting  in  a
-- dimensioned fuzzy number. 
--
-- Returns :
--
--    Left * Right
--
-- Exceptions :
--
--    Constraint_Error - Unit power overflow
--    Unit_Error       - Incompatible units
--
   function "*" (Left, Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "*" (Left : Fuzzy_Measure; Right : Number)
      return Fuzzy_Measure;
   function "*" (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Measure;
   function "*" (Left : Fuzzy_Measure; Right : Interval)
      return Fuzzy_Measure;
   function "*" (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Measure;
   function "*" (Left : Fuzzy_Measure; Right : Fuzzy_Float)
      return Fuzzy_Measure;
   function "*" (Left : Number; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "*" (Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "*" (Left : Interval; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "*" (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "*" (Left : Fuzzy_Float; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "*" (Left : Measure; Right : Fuzzy_Float)
      return Fuzzy_Measure;
   function "*" (Left : Interval_Measure; Right : Fuzzy_Float)
      return Fuzzy_Measure;
   function "*" (Left : Fuzzy_Float; Right : Measure)
      return Fuzzy_Measure;
   function "*" (Left : Fuzzy_Float; Right : Interval_Measure)
      return Fuzzy_Measure;
--
-- / -- Division
--
--    Left  - First argument
--    Right - Second argument
--
-- One  of  the  argument  is  a  fuzzy measure another is either of the
-- following types:  fuzzy  measure,  fuzzy  number,  number,  interval,
-- interval  measure.  Additionally division is defined on a dimensioned
-- number or interval and a fuzzy  number  resulting  in  a  dimensioned
-- fuzzy number.  
--
-- Returns :
--
--    Left / Right
--
-- Exceptions :
--
--    Constraint_Error - Unit power overflow
--    Unit_Error       - Incompatible units
--
   function "/" (Left, Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "/" (Left : Fuzzy_Measure; Right : Number)
      return Fuzzy_Measure;
   function "/" (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Measure;
   function "/" (Left : Fuzzy_Measure; Right : Interval)
      return Fuzzy_Measure;
   function "/" (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Measure;
   function "/" (Left : Fuzzy_Measure; Right : Fuzzy_Float)
      return Fuzzy_Measure;
   function "/" (Left : Number; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "/" (Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "/" (Left : Interval; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "/" (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "/" (Left : Fuzzy_Float; Right : Fuzzy_Measure)
      return Fuzzy_Measure;
   function "/" (Left : Measure; Right : Fuzzy_Float)
      return Fuzzy_Measure;
   function "/" (Left : Interval_Measure; Right : Fuzzy_Float)
      return Fuzzy_Measure;
   function "/" (Left : Fuzzy_Float; Right : Measure)
      return Fuzzy_Measure;
   function "/" (Left : Fuzzy_Float; Right : Interval_Measure)
      return Fuzzy_Measure;
--
-- ** -- Exponentiation
--
--    Left  - First argument
--    Right - Second argument (natural power)
--
-- Returns :
--
--    Left ** Right
--
-- Exceptions :
--
--    Constraint_Error - Unit power overflow
--
   function "**" (Left : Fuzzy_Measure; Right : Natural)
      return Fuzzy_Measure;
--
-- >, >=, =, /=, <, <= -- Comparisons
--
--    Left  - A measure / interval measure / fuzzy measure
--    Right - A measure / interval measure / fuzzy measure
--
-- Comparisons are allowed for differently shifted arguments. One of the
-- arguments can be measure or plain interval. 
--
-- Returns :
--
--    The result of comparison
--
-- Exceptions :
--
--    Unit_Error - Incompatible units
--
   function ">" (Left, Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function ">" (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Boolean;
   function ">" (Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function ">" (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Boolean;
   function ">" (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;

   function ">=" (Left, Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function ">=" (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Boolean;
   function ">=" (Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function ">=" (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Boolean;
   function ">=" (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;

   function "=" (Left, Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function "=" (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Boolean;
   function "=" ( Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function "=" (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Boolean;
   function "=" ( Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;

   function "/=" (Left, Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function "/=" (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Boolean;
   function "/=" (Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function "/=" (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Boolean;
   function "/=" (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;

   function "<" (Left, Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function "<" (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Boolean;
   function "<" (Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function "<" (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Boolean;
   function "<" (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;

   function "<=" (Left, Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function "<=" (Left : Fuzzy_Measure; Right : Measure)
      return Fuzzy_Boolean;
   function "<=" (Left : Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;
   function "<=" (Left : Fuzzy_Measure; Right : Interval_Measure)
      return Fuzzy_Boolean;
   function "<=" (Left : Interval_Measure; Right : Fuzzy_Measure)
      return Fuzzy_Boolean;

end Fuzzy.Measures;
