--                                                                    --
--  package Measures_Irregular      Copyright (c)  Dmitry A. Kazakov  --
--  Interface                                      Luebeck            --
--                                                 Spring, 2000       --
--                                                                    --
--                                Last revision :  22:14 12 Sep 2005  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
-- This package provides some of the great variety  of  irregular  units
-- still  used  parallel to SI ones. It also contains some SI-compatible
-- units,  like  liter  and metric ton. They are defined as constants of
-- the  type  Measure.  The  package  is  generic.  The  formal  generic
-- parameter  is  an  instance  of  the  package  Measures_Derived.   To
-- instantiate this package one should write something like this: 
--
-- with Measures;
--    ...
--    --
--    -- Instantiate Measures with type Float as the parameter
--    --
--    package Float_Measures is new Measures (Float);
--    --
--    -- Instantiate Measures_Derived
--    --
--    package Float_Measures_Derived is
--       new Measures_Derived (Float_Measures);
--    --
--    -- Instantiate Measures_Irregular
--    --
--    package Float_Measures_Irregular is
--       new Measures_Irregular (Float_Measures_Derived);
--
-- The package Float_Measures_Irregular is a non-generic version of  the
-- package based on the type Float.
--
with Measures_Derived;

generic
   with package Derived_Measures is new Measures_Derived (<>);
package Measures_Irregular is
   use Derived_Measures;
   use Derived_Measures.Measures;

   pi   : constant := 3.14159_26535_89793_23846_26433_83279_5;
   ln10 : constant := 2.30258_50929_94045_68401_79914_54684_4; 
   --
   -- Some of irregular measurement units
   --
   ft   : aliased constant Measure := 0.3048 * m;      -- Foot

   acre : aliased constant Measure := 4050.0 * m**2;   -- Acre
   ang  : aliased constant Measure := 1.0e-10 * m;     -- Angstroem
   are  : aliased constant Measure := 100.0 * m**2;    -- Are
   atm  : aliased constant Measure := 1.01325e+5 * Pa; -- Atmosphere

   B          : aliased constant Measure := Np * ln10 * 0.5; -- Bel
   BTU        : aliased constant Measure := 1055.06 * J;     -- British Termal Unit
   Bq         : aliased constant Measure := 1.0 / s;         -- Becquerel
   bar        : aliased constant Measure := 100000.0 * Pa;   -- Bar
   barleycorn : aliased constant Measure := Ft / 36.0;       -- Barleycorn (inch/12)
   barn       : aliased constant Measure := 1.0e-28 * m**2;  -- Barn
   barrel     : aliased constant Measure := 1.5898 * m**3;   -- Barrel

   Ci      : aliased constant Measure := 3.7e+10 * Bq;       -- Curie
   cal     : aliased constant Measure := 41868.0 * J;        -- Calorie
   candle  : aliased constant Measure := 12.566 * cd;        -- Candle
   carat   : aliased constant Measure := 200.0e-6 * kg;      -- Carat
   Celsius : aliased constant Measure := K and 273.15;       -- Celsius
   ch      : aliased constant Measure := 66.0 * Ft;          -- Chain
   cubit   : aliased constant Measure := 18.0 * Ft / 12.0;   -- Cubit (18 inches)

   d       : aliased constant Measure := 24.0 * 3600.0 * s;  -- Day
   dB      : aliased constant Measure := 10.0 * B;           -- Decibel
   degree  : aliased constant Measure := pi * rad / 180.0;   -- Degree (angle)
   dyn     : aliased constant Measure := 1.0e-5 * N;         -- Dyn

   eV      : aliased constant Measure := 1.60218e-19 * J; -- Electronvolt
   ell     : aliased constant Measure := 1.14 * m;        -- Ell
   erg     : aliased constant Measure := 1.0e-7 * J;      -- Erg

   Fahrenheit : aliased constant Measure := K * 5.0 / 9.0 and 255.372222; -- Fahrenheit
   fathom     : aliased constant Measure := Ft * 6.0;      -- Fathom
   finger     : aliased constant Measure := 0.114 * m;     -- Finger
   fpm        : aliased constant Measure := ft / 60.0 * s; -- Feet per minute
   fps        : aliased constant Measure := ft / s;        -- Feet per second 
   fur        : aliased constant Measure := Ft * 660.0;    -- Furlong

   G      : aliased constant Measure := 1.0e-4 * kg / (A * s**2); -- Gauss
   gal    : aliased constant Measure := 0.03785 * m**3;           -- Gallon
   grain  : aliased constant Measure := 64.8e-6 * kg;             -- Grain
   gram   : aliased constant Measure := kg / 1000.0;              -- Gram

   h          : aliased constant Measure := 3600.0 * s;     -- Hour
   hand       : aliased constant Measure := Ft / 3.0;       -- Hand
   hectare    : aliased constant Measure := 10000.0 * m**2; -- Hectare
   hp         : aliased constant Measure := W / 0.001341;   -- Horsepower

   INM    : aliased constant Measure := 1852.0 * m; -- International Nautical Mile
   inch   : aliased constant Measure := ft / 12.0;  -- Inch

   kcal   : aliased constant Measure := 1000.0 * cal;    -- Kilocalorie 
   kgf    : aliased constant Measure := 9.80665 * N;     -- Kilogram-force
   knot   : aliased constant Measure := 1852.0 * m / h;  -- Knot

   L      : aliased constant Measure := 0.001 * m**3;    -- Liter
   lb     : aliased constant Measure := 0.45359237 * kg; -- Pound
   line   : aliased constant Measure := Inch / 12.0;     -- Line
   link   : aliased constant Measure := Ft;              -- Link
   ly     : aliased constant Measure := 9.46053e+15 * m; -- Light year

   mi         : aliased constant Measure := 1609.0 * m;     -- Land mile
   min        : aliased constant Measure := 60.0 * s;       -- Minute
   min_of_arc : aliased constant Measure := degree / 60.0;  -- Minute of arc
   mpg        : aliased constant Measure := mi / gal;       -- Miles per gallon
   mph        : aliased constant Measure := mi / h;         -- Miles per hour
   mps        : aliased constant Measure := mi / s;         -- Miles per second

   nail   : aliased constant Measure := Inch * 2.5;         -- Nail

   Oe : aliased constant Measure := 79.577472 * A / m;      -- Oersted
   oz : aliased constant Measure := Lb / 16.0;              -- Ounce

   pace    : aliased constant Measure := 0.762 * m;         -- Pace
   pc      : aliased constant Measure := 3.08568e+16 * m;   -- Parsec
   percent : aliased constant Measure := 0.01 * Np;         -- Percent
   point   : aliased constant Measure := Inch / 72.0;       -- Point
   ppb     : aliased constant Measure := Np / 1.0e+09;      -- Parts per billion
   ppm     : aliased constant Measure := Np / 1.0e+06;      -- Parts per million
   ppt     : aliased constant Measure := Np / 1.0e+12;      -- Parts per trillion
   psi     : aliased constant Measure := Pa * 6894.7;       -- Pounds per square inch
   pt      : aliased constant Measure := 0.4731765 * L;     -- Pint

   qt   : aliased constant Measure := 0.946 * L;            -- Quart

   R    : aliased constant Measure := 2.58e-4 * C / kg;     -- Roentgen
   rd   : aliased constant Measure := 5.5 * 3.0 * Ft;       -- Rod
   rood : aliased constant Measure := 1010.0 * m**2;        -- Rood
   rpm  : aliased constant Measure := 360.0 * degree / min; -- Revolutions per minute
   rps  : aliased constant Measure := 360.0 * degree / s;   -- Revolutions per second

   sec_of_arc : aliased constant Measure := degree / 3600.0;   -- Second of arc
   span       : aliased constant Measure := 9.0 * Inch;        -- Span

   t            : aliased constant Measure := 1000.0 * kg;     -- Metric ton
   tablespoon   : aliased constant Measure := 0.0142 * L;      -- Tablespoon
   teaspoon     : aliased constant Measure := 0.0044 * L;      -- Teaspoon
   torr         : aliased constant Measure := 133.3226 * Pa;   -- Torr (mm Hg)
   township     : aliased constant Measure := 6.0 * mi**2;     -- Township

   u  : aliased constant Measure := 1.66054e-27 * kg;    -- Unified atomic mass
   ua : aliased constant Measure := 1.49598e+11 * m;     -- Astronomical unit

   wineglass : aliased constant Measure := 0.056 * L;    -- Wineglass

   yd   : aliased constant Measure := 3.0 * Ft;          -- Yard
   year : aliased constant Measure := 3.1556926e+7 * s;  -- Year
end Measures_Irregular;
