--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Name_Tables                                 Luebeck            --
--  Interface                                      Spring, 2003       --
--                                                                    --
--                                Last revision :  16:12 16 Apr 2005  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
--  This package  instantiates  table  packages  required  for  sets  of
--  linguistic variables. 
--
with Ada.Strings.Maps;            use Ada.Strings.Maps;
with Ada.Strings.Maps.Constants;  use Ada.Strings.Maps.Constants;
with Tables.Names;

package Name_Tables is
   pragma Elaborate_Body (Name_Tables);
--
-- The following variables determine valid names:
--
-- (o)  A valid  name  starts  with  one  of  characters  from  the  set
--      specified by the variable Capitals; 
-- (o)  The rest of the name can contain any number of  characters  from
--      the set Name_Body. In this set, if space  is  a  valid character
--      then  when  matched  it matches any non-empty sequence of spaces
--      and tabs. Letters are matched case insensitive; 
-- (o)  Characters from the Singletons when allowed in the name body and
--      appear there shall be followed by a non-singleton character. For
--      example, if '_' is a singleton then "a__b" is not a valid  name.
--      A singleton may not appear in the end of a name. So  "a_"  would
--      be also an invalid name;  
-- (o)  When matched a valid name may not be  followed  by  any  of  the
--      characters  from  Non_Breaks.  This is usually Name_Body without
--      Singletons. This rule overrides all others  if  they  contradict
--      it.  
--
   Capitals   : Character_Set := Letter_Set;
   Singletons : Character_Set := To_Set (" _");
   Name_Body  : Character_Set := Alphanumeric_Set or Singletons;
   Non_Breaks : Character_Set := Alphanumeric_Set;
--
-- Check_Matched -- Check name margin
--
--    Source  - The source string
--    Pointer - Pointer to the character following the name
--
-- Returns :
--
--    True if Source (Pointer) is not in Non_Breaks
--
   function Check_Matched (Source : String; Pointer : Integer)
      return Boolean;
--
-- Check_Name -- Check spelling of a name
--
--    Name - To be checked
--
-- See valid name description above.
--
-- Exceptions :
--
--    Constraint_Error - Invalid name
--
   procedure Check_Name (Name : String);
--
-- String_Maps -- Tables of strings
--
   package String_Maps is new Tables (Integer);
--
-- Name_Maps -- Tables of identifiers
--
   package Name_Maps is
      new String_Maps.Names (Check_Name, Check_Matched);
--
-- Get_Name -- Get a valid name from string
--
--    Source  - The string
--    Pointer - Source (Pointer.all) is the character to start from
--
-- This  function  reconizes  a  valid  (see Check_Name) name in Source.
-- Spaces and tabs following  the  name  do  not  influence  success  or
-- failure of matching. On success Pointer.all is advanced to the  first
-- character following the name.
--
-- Returns :
--
--    The name
--
-- Exceptions :
--
--    Constraint_Error - Illegal name
--    End_Error        - Nothing matched
--    Layout_Error     - Pointer.all is not in Source'First..Source'Last
--
   function Get_Name (Source : String; Pointer : access Integer)
      return String;

end Name_Tables;
