--                                                                    --
--  package Persistent.Handle       Copyright (c)  Dmitry A. Kazakov  --
--  Interface                                      Luebeck            --
--                                                 Winter, 2002       --
--                                                                    --
--                                Last revision :  12:32 10 Jun 2003  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
--  This  package  defines  the  type  Storage_Handle which is used to
--  access data base objects. 
--
package Persistent.Handle is
   pragma Elaborate_Body (Persistent.Handle);

   type Storage_Handle is private;
--
-- Get -- Get a persistent object by name
--
--    Storage - A handle to
--    Name    - Name of the object
--
-- This function searches for the specified object by its name.  If  the
-- object  is already available a handle to it is returned. Otherwise it
-- first is restored from the persistent storage. 
--
-- Returns :
--
--    A handle to the object
--
-- Exceptions :
--
--    Constraint_Error - Invalid handle
--    Data_Error       - Inconsistent storage
--    End_Error        - No such object
--
   function Get
            (  Storage : Storage_Handle;
               Name    : Wide_String
            )  return Deposit_Handle;
   pragma Inline (Get);
--
-- Get_List -- List of persistent objects
--
--    Storage     - A handle to
--    Prefix      - Of names to get
--    Suffix      - Of the names
--    Equivalence - Of the characters when compared
--
-- This function returns a complete list of all named objects persistent
-- in Storage. The list does not include anonymous  persistent  objects.
-- Only names starting with Prefix and ending with Suffix are  returned.
-- When  names  are compared two characters are considered same if their
-- corresponding  values  in  the  mapping Equivalence are same. Observe
-- that Prefix may not overlap Suffix when matched.  So  if  Prefix="AB"
-- and Suffix="BC", then "ABC" does not fit, but "ABBC" does.
--
-- Returns :
--
--    Names of the objects in the data base
--
-- Exceptions :
--
--    Constraint_Error - Invalid handle
--    Data_Error       - Inconsistent storage
--
   function Get_List
            (  Storage     : Storage_Handle;
               Prefix      : Wide_String := "";
               Suffix      : Wide_String := "";
               Equivalence : Wide_Character_Mapping := Identity
            )  return Catalogue.Set;
   pragma Inline (Get_List);
--
-- Get_Name -- Of a persistent object
--
--    Storage - A handle to
--    Object  - A handle to
--
-- Returns :
--
--    The object name in Storage
--
-- Exceptions :
--
--    Constraint_Error - Invalid handle or not a persistent object
--    Data_Error       - Inconsistent storage
--    Name_Error       - Object is anonymous
--
   function Get_Name
            (  Storage : Storage_Handle;
               Object  : Deposit_Handle
            )  return Wide_String;
   pragma Inline (Get_Name);
--
-- Invalidate -- Detach handle from the object
--
--    Storage - A handle to
--
-- This procedure makes handle pointing to nothing. If it was  the  last
-- reference to the object, the latter is destroyed. 
--
   procedure Invalidate (Storage : in out Storage_Handle);
   pragma Inline (Invalidate);
--
-- Is_In -- Check if an object persists in the storage
--
--    Storage       - A handle to
--    Name / Object - Name of or a handle to a persistent object
--
-- This function checks whether Name / Object persists  in  Storage.  If
-- Object is not a valid handle, the result is False.
--
-- Returns :
--
--    True if Object persists in Storage 
--
-- Exceptions :
--
--    Constraint_Error - Storage is not a valid handle
--    Data_Error       - Inconsistent storage 
--
   function Is_In
            (  Storage : Storage_Handle;
               Object  : Deposit_Handle
            )  return Boolean;
   function Is_In
            (  Storage : Storage_Handle;
               Name    : Wide_String
            )  return Boolean;
   pragma Inline (Is_In);
--
-- Is_Valid -- Check if the handle is associated with an object
--
--    Storage - The handle
--
-- Returns :
--
--    True if the handle can de dereferenced
--
   function Is_Valid (Storage : Storage_Handle) return Boolean;
   pragma Inline (Is_Valid);
--
-- Ptr -- Get the pointer to the object by a handle
--
--    Storage - The handle
--
-- Returns :
--
--    The referenced object
--
   function Ptr (Storage : Storage_Handle) return Storage_Object_Ptr;
   pragma Inline (Ptr);
--
-- Put -- Make an object persistent
--
--    Storage - A handle to
--    Object  - A handle to the object to store in Storage
--  [ Name  ] - The name of the object being stored
--
-- This procedure is used to store Object in Storage. The parameter Name
-- specifies the object name there. When omitted the  object  is  stored
-- anonymous. Anonymous persistent objects are collected  when  no  more
-- used. When Object already persists in Storage and Name is  specified,
-- then  it  is  checked  that  it is same. If this check fails, Name is
-- illegal,  or  conflicts with the name of another object Name_Error is
-- propagated. When name is not specified no check is made.
--
-- Exceptions :
--
--    Constraint_Error - Invalid handle
--    Data_Error       - Inconsistent storage
--    Name_Error       - Name conflict
--
   procedure Put
             (  Storage : in out Storage_Handle;
                Object  : in out Deposit_Handle;
                Name    : Wide_String
             );
   procedure Put
             (  Storage : in out Storage_Handle;
                Object  : in out Deposit_Handle
             );
   pragma Inline (Put);
--
-- Ref -- Get handle to a persistent storage object
--
--    Storage - The object
--
-- Returns :
--
--    Handle to the object
--
   function Ref (Storage : Storage_Object_Ptr) return Storage_Handle;
   pragma Inline (Ref);
--
-- Rename -- A persistent objects
--
--    Storage  - A handle to
--    Object   - The object to be renamed (a handle to or the name of)
--    New_Name - The new object name
--
-- This procedure changes the name of the object specified by either its
-- old name or by a handle to it.  When  renamed  object  was  anonymous
-- before renaming it becomes a named one. When  Object  is  an  invalid
-- handle or does not refer a persistent object then Constraint_Error is
-- propagated.  End_Error is propagated when Old_Name does not refer any
-- persistent object. Name_Error indicates  an  illegal  or  conflicting
-- name.  
--
-- Exceptions :
--
--    Constraint_Error - Invalid handle or not a persistent object
--    Data_Error       - Inconsistent storage
--    End_Error        - No such object
--    Name_Error       - Name conflict, there is another object named so
--
   procedure Rename
             (  Storage  : in out Storage_Handle;
                Object   : Wide_String;
                New_Name : Wide_String
             );
   procedure Rename
             (  Storage  : in out Storage_Handle;
                Object   : in out Deposit_Handle;
                New_Name : Wide_String
             );
   pragma Inline (Rename);
--
-- Unname -- A persistent objects
--
--    Storage - A handle to
--    Object  - The object to be unnamed (a handle to or the name of)
--
-- This procedure  makes  the  object  anonymous.  Unnamed  objects  are
-- automatically  deleted  when  no  more in use. Nothing happens if the
-- object is already unnamed or does not persist in Storage.  Note  that
-- anonymous  objects  are  not  deleted  as  long  as  they have memory
-- resident counterparts. Observe a difference between Unname and Delete
-- called on an object handle. Delete requests object deletion from both
-- memory  and persistent storage. Unname does it for persistent storage
-- only.  Both  may  have  no immediate effect if the object is still in
-- use.
--
-- Exceptions :
--
--    Constraint_Error - Invalid handle (Storage)
--    Data_Error       - Inconsistent storage
--
   procedure Unname
             (  Storage : in out Storage_Handle;
                Object  : Wide_String
             );
   procedure Unname
             (  Storage : in out Storage_Handle;
                Object  : in out Deposit_Handle
             );
   pragma Inline (Unname);

private
   type Storage_Handle is new Handles.Handle with null record;
      
end Persistent.Handle;
