--                                                                    --
--  package Persistent              Copyright (c)  Dmitry A. Kazakov  --
--  Interface                                      Luebeck            --
--                                                 Winter, 2002       --
--                                                                    --
--                                Last revision :  12:32 10 Jun 2003  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
--  This package defines an interface to an abstract persistent storage.
--  The abstract type Storage_Object describes the interface. Persistent
--  object are identified by their names. Additionally anonymous objects
--  can be created and deleted as required by the named  ones.  When  an
--  object depends on some other objects,  then  when  stored  into  the
--  storage,  the  referred  objects  are  stored as well if they do not
--  already persist there. In such case these objects will be anonymous.
--  Anonymous persistent objects are subject of garbage collection.  The
--  way of collection is determined by the implementation. Named objects
--  are  deleted only on request. For this Object.Archived.Delete can be
--  applied to the object's handle. If  the  object  cannot  be  deleted
--  immediatelly it becomes anonymous for later  collection.  Persistent
--  storage  interfaces  are itself objects and are a subject of garbage
--  collection  as  well.  The  object  names  are   in   Unicode.   The
--  implementation should provide encoding when the  persistent  storage
--  does not natively support Unicode. 
--
with Ada.Strings.Wide_Unbounded;  use Ada.Strings.Wide_Unbounded;
with Ada.Strings.Wide_Maps;       use Ada.Strings.Wide_Maps;
with Deposit_Handles;             use Deposit_Handles;

with Generic_Set;
with Object.Handle;

package Persistent is
--
-- Storage_Object -- Represents a persistent storage interface
--
   type Storage_Object is abstract new Object.Entity with private;
   type Storage_Object_Ptr is access Storage_Object'Class;
   for Storage_Object_Ptr'Storage_Pool
      use Object.Entity_Ptr'Storage_Pool;
   subtype Deposit_Handle is Deposit_Handles.Handle;
--
-- Catalogue -- List of objects names
--
   package Catalogue is
      new Generic_Set
          (  Object_Type  => Unbounded_Wide_String,
             Null_Element => Null_Unbounded_Wide_String
          );
--
-- Get -- Get a persistent object by name
--
--    Storage - A pesistent storage object (a pointer to)
--    Name    - Of the object
--
-- This function returns a handle to a persistent object by its name. An
-- implementation should  first  check  if  the  the  persistent  object
-- already has a memory-resident counterpart. Otherwise it should create
-- one from the persistent storage.
--
-- Returns :
--
--    A handle to the object
--
-- Exceptions :
--
--    Data_Error - Inconsistent data base
--    End_Error  - No such object
--
   function Get
            (  Storage : access Storage_Object;
               Name    : Wide_String
            )  return Deposit_Handle is abstract;
--
-- Get_List -- List of persistent objects
--
--    Storage     - A pesistent storage object (a pointer to)
--    Prefix      - Of names to get
--    Suffix      - Of the names
--    Equivalence - Of the characters when compared
--
-- This function returns a complete list of all named objects persistent
-- in Storage. The list does not include anonymous  persistent  objects.
-- Only names starting with Prefix and ending with Suffix are  returned.
-- When  names  are compared two characters are considered same if their
-- corresponding  values in the mapping Equivalence are same. Prefix and
-- Suffix may not overlap when matched.
--
-- Returns :
--
--    Names of the objects in the data base
--
-- Exceptions :
--
--    Data_Error - Inconsistent data base
--
   function Get_List
            (  Storage     : access Storage_Object;
               Prefix      : Wide_String := "";
               Suffix      : Wide_String := "";
               Equivalence : Wide_Character_Mapping := Identity
            )  return Catalogue.Set is abstract;
--
-- Get_Name -- Of a persistent object
--
--    Storage - A pesistent storage object (a pointer to)
--    Object  - A handle to
--
-- This function returns the object's name in Storage.  Constraint_Error
-- is propagated when Object does not persist in Storage. When  it  does
-- but is anonymous, then Name_Error is propagated. 
--
-- Returns :
--
--    The object's name in Storage
--
-- Exceptions :
--
--    Constraint_Error - Invalid handle or not a persistent object
--    Data_Error       - Inconsistent data base
--    Name_Error       - Object is anonymous
--
   function Get_Name
            (  Storage : access Storage_Object;
               Object  : Deposit_Handle
            )  return Wide_String is abstract;
--
-- Is_In -- Check if an object is persistent
--
--    Storage       - A pesistent storage object (a pointer to)
--    Name / Object - The name of or a handle to an object there
--
-- These  functions  check  whether  an  object persists in Storage. The
-- object can be identified either by its name or by  a  handle  to  it.
-- When Object is not a valid handle the result is False.  
--
-- Returns :
--
--    True if the object persists in Storage 
--
-- Exceptions :
--
--    Data_Error - Inconsistent data base
--
   function Is_In
            (  Storage : access Storage_Object;
               Name    : Wide_String
            )  return Boolean is abstract;
   function Is_In
            (  Storage : access Storage_Object;
               Object  : Deposit_Handle
            )  return Boolean is abstract;
--
-- Put -- Make an object persistent
--
--    Storage - A pesistent storage
--  [ Name  ] - The name of the object being stored
--    Object  - A handle to the object to store in Storage
--
-- These  procedures  are used to store Object in Storage. The parameter
-- Name specifies the object's name there. When omitted  the  object  is
-- stored as anonymous. Anonymous persistent objects are collected  when
-- not used, but not before its  memory-resident  counterpart  vanishes.
-- When  Object  already persists in Storage and Name is specified, then
-- it is checked to be same. If this  check  fails,  Name  is  empty  or
-- illegal,  or  conflicts with the name of another object Name_Error is
-- propagated. When name is not specified, no check is made.  
--
-- Exceptions :
--
--    Constraint_Error - Invalid handle
--    Data_Error       - Inconsistent data base
--    Name_Error       - Name conflict, illegal or empty name
--
   procedure Put
             (  Storage : in out Storage_Object;
                Object  : in out Deposit_Handle;
                Name    : Wide_String
             )  is abstract;
   procedure Put
             (  Storage : in out Storage_Object;
                Object  : in out Deposit_Handle
             )  is abstract;
--
-- Rename -- A persistent object
--
--    Storage           - A pesistent storage object
--    Old_Name / Object - Object to be renamed
--    New_Name          - The new name
--
-- These procedures change the name of an object specified by either its
-- old name or by a handle to it.  When  renamed  object  was  anonymous
-- before renaming it becomes a named one. When  Object  is  an  invalid
-- handle or does not refer a persistent object then Constraint_Error is
-- propagated.  End_Error is propagated when Old_Name does not refer any
-- persistent object. Name_Error indicates  an  illegal  or  conflicting
-- name. --
-- Exceptions :
--
--    Constraint_Error - Invalid handle or not a persistent object
--    Data_Error       - Inconsistent storage
--    End_Error        - No such object
--    Name_Error       - Name conflict, there is another object named so
--
   procedure Rename
             (  Storage  : in out Storage_Object;
                Old_Name : Wide_String;
                New_Name : Wide_String
             )  is abstract;
   procedure Rename
             (  Storage  : in out Storage_Object;
                Object   : in out Deposit_Handle;
                New_Name : Wide_String
             )  is abstract;
--
-- Unname -- A persistent object
--
--    Storage - A pesistent storage object
--    Object  - To be unnamed, a handle to or the name of
--
-- These  procedures  make  object  anonymous.   Unnamed   objects   are
-- automatically  deleted  when  no  more in use. Nothing happens if the
-- object is already unnamed. Nothing also happens if Object is invalid;
-- not a handle to a persistent object or  does  not  exist.  Note  that
-- anonymous  objects  are  not  deleted  as  long  as  they have memory
-- resident counterparts.  Observe difference between  Unname and Delete
-- called on an object handle. Delete requests object deletion from both
-- memory  and persistent storage. Unname does it for persistent storage
-- only.  Both  may  have  no immediate effect if the object is still in
-- use.
--
-- Exceptions :
--
--    Data_Error - Inconsistent storage
--
   procedure Unname
             (  Storage : in out Storage_Object;
                Object  : Wide_String
             )  is abstract;
   procedure Unname
             (  Storage : in out Storage_Object;
                Object  : in out Deposit_Handle
             )  is abstract;

private
   type Storage_Object is abstract new Object.Entity with null record;
--
-- Handles -- Handles to persistent objects
--
   package Handles is
      new Object.Handle (Storage_Object, Storage_Object_Ptr);
end Persistent;
